# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Fri Jan  7 11:33:47 2022

Portal truss: two columns and a truss

TODO (17.1.2023)
    
    1. Liitokset pilareihin: epäkeskisyys alapaarteella ja yläpaarteella
    2. FEM-malli EN 1993 -epäkeskisyyksillä.
    
    Nyt create_truss tekee ristikon liitokset. Kun alapaarre tulee kiinni
    pilariin ja ensimmäinen diagonaali lähtee alapaarteelta yläpaarteelle, ei
    yläpaarteen ja pilarin välille tule paarresauvalle liitosta.
    
    Ristikon ja pilarin liitoksille täytyy luoda omat olionsa.
    
    ** Yläpaarteella on kaksi vaihtoehtoa
    a) yläpaarre tulee pilarin päälle. Tällöin ei tarvita epäkeskisyyttä, ja
        FEM-solmut alkavat paarteen ja pilarin keskilinjojen leikkauspisteestä.
    
    b) yläpaarre tulee pilarin kylkeen. Tällöin FEM-solmut tehdään pilarin päähän
        ja kylkeen. Pilarin kylkeen tulevan solmun kohdalle tulee nivel
        
    Jos ensimmäinen diagonaali lähtee yläpaarteelta alapaarteelle, on diagionaalin
    pää tuotava riittävän etäälle pilarin kyljestä.


    ** Jos alapaarre tulee kiinni pilariin, tulee se pilarin kylkeen. Tällöin
    menetellään kuten yläpaarteen kanssa, eli solmut tehdään pilarin keskilinjan
    ja laipan reunan kohdille. Alapaarteelta lähtevä diagonaali on tuotava
    sopivan etäisyyden päähän pilarin reunasta.    
    
    raami_plane_truss-tiedostossa TopChord -luokan generate_elements ei toimi, jos
    eccentricity on 'en1993', koska sauvalla on vain yksi liitos

@author: kmela
"""

from copy import copy
import matplotlib.pyplot as plt
import numpy as np

from metku.raami.raami import Raami
from metku.raami.frame_member import FrameMember, SteelFrameMember, MultiSpanSteelMember, MemberGroup
from metku.raami.frame_node import FrameNode
from metku.raami.frame_loads import PointLoad, LineLoad, LoadIDs
from metku.raami.frame_supports import XYHingedSupport, FixedSupport
from metku.raami.raami_plane_truss import SlopedTruss, TubularYJoint, TubularKGapJoint, TrussBrace

from metku.framefem.elements.ebbeam import EBBeam

from metku.raami.exports import AbaqusOptions

from metku.sections.steel.ISection import HEA, HEB, IPE
from metku.sections.steel.RHS import RHS, SHS

class PortalTruss(Raami):
    """ Class for portal frame trusses """


    def __init__(self,span=20000,height_eaves=10000,column=None,truss=None,name="Portal Truss"):
        
        super().__init__(name)
                
        self.L = span
        self.height = height_eaves
        self.column_profile = column
        self.truss = truss
        self.columns = []
    
        self.bottom_chord_to_column = False
    
        # Create nodes for column bases
        self.add(FrameNode([0,0]))
        self.add(FrameNode([self.L,0]))
    
        self.column_joints = []
    
    def __repr__(self):
        
        return f"{self.__name__}: height = {self.height:.2f} mm, span = {self.L:.2f} mm"
    
    @property
    def col_length(self):
        """ total length of the column """
        return self.height
    
    @property
    def col_profile(self):
        """ cross-section of the profiles """
        if len(self.columns) == 0:
            p = self.column_profile
        else:
            p = self.columns[0].cross_section
        return p
    
    @col_profile.setter
    def col_profile(self,val):
        """ set cross-section of the profiles """
        self.column_profile = val
        
        for col in self.columns:
            col.cross_section = val    
        
    def weight_detailed(self,verb=False):
        """ Returns detailed weight 
            The weight of truss and columns are separated.
        """
        
        Wtop, Wbottom, Wbraces = self.truss.weight_detailed(False)
        
        Wcols = 2*self.columns[0].weight()
        
        Wtruss = Wtop + Wbottom + Wbraces
        Wtot = Wtruss + Wcols
        
        if verb:
            print(f"Portal Truss weight: {Wtot:.2f} kg")
            print(f"Weight of columns (total): {Wcols:.2f} kg ({Wcols/Wtot*100:.1f}%)")            
            print(f"Truss weight: {Wtruss:.2f} kg ({Wtruss/Wtot*100:.1f}%)")            
            print(f"Top Chord: {Wtop:.2f} kg")
            print(f"Bottom Chord: {Wbottom:.2f} kg")
            print(f"Braces: {Wbraces:.2f} kg")
        
        return Wtop, Wbottom, Wbraces
    
    def ridge_height(self):
        """ Height at ridge level """
        return self.height + self.truss.height_of_slope_left()
    
    def add(self, this):
        """
        Adds object to portal truss
        :param this: object to be added
        :return:
        """
        # COLUMN
        if (isinstance(this, SteelFrameMember) or isinstance(this, MultiSpanSteelMember)) and this.mtype == 'column':
            self.columns.append(this)
            this.mem_id = len(self.members)
            self.members["COL" + str(this.mem_id)] = this
            this.frame = self
            
            if isinstance(this, MultiSpanSteelMember):
                for mem in this.members:
                    mem.frame = self
        else:
            super().add(this)
    
    def create_columns(self,profile=HEA(260),hinges=[False,False], lcr=[1.0,1.0], nel=[4,2]):
        """ creates columns """        
        
        if not self.truss is None:
            # Truss has already been created
            
            if self.bottom_chord_to_column:
                # Bottom chord is connected to the column
                self.add(MultiSpanSteelMember([self.nodes[0],self.truss.bottom_nodes[0],self.truss.top_nodes[0]],
                                             profile,mem_type='column',nel=nel,hinges=hinges,lcr=lcr,sway=True))                    
                self.add(MultiSpanSteelMember([self.nodes[1],self.truss.bottom_nodes[-1],self.truss.top_nodes[-1]],profile,
                                                          mem_type='column',nel=nel,hinges=hinges,lcr=lcr,sway=True))
            else:
                if isinstance(nel,list):
                    nel = sum(nel)
                self.add(SteelFrameMember([self.nodes[0],self.truss.top_nodes[0]],profile,
                                                      mem_type='column',nel=nel,hinges=hinges,lcr=lcr,sway=True))
            
                self.add(SteelFrameMember([self.nodes[1],self.truss.top_nodes[-1]],profile,
                                                      mem_type='column',nel=nel,hinges=hinges,lcr=lcr,sway=True))
        
        col_group = MemberGroup(self.columns,'columns')
        self.member_groups['columns'] = col_group
            
    
    def create_truss(self, topology='K', ndiv=4, truss_type='simple', **kwargs):
        """ Makes the truss based on user data """
        
        if truss_type == 'simple':
            # Simple truss means that the bottom chord is not connected to the
            # column.
            try:
                dx1 = kwargs['dx']
                dx2 = kwargs['dx']
            except:
                dx1 = 1500
                dx2 = 1500            
        else:
            # Non-simple truss means that the bottom chord is connected to the column
            dx1 = 0
            dx2 = 0
        
        try:
            first = kwargs['first_diagonal_up']
        except:
            first = True
        
        try:
            nel_chord = kwargs['nel_chord']
        except:
            nel_chord = 4
        
        try:
            nel_brace = kwargs['nel_brace']            
        except:
            nel_brace = 1
        
        if dx1 == 0:
            self.bottom_chord_to_column = True
                
        
        self.truss = SlopedTruss(L1=0.5*self.L,L2=0.5*self.L,dx1=dx1,dx2=dx2,**kwargs)
        
        if nel_brace > 1:
            self.truss.braces_as_beams = True
        
        self.truss.origin = (0,self.height)
        self.truss.generate_topology(topology,ndiv,first_diagonal_up=first,edge_verticals=False,nel_chord=nel_chord,nel_brace=nel_brace)
        self.truss.generate_joints()
        self.truss.fem = self.fem
        self.truss.symmetry()
        
        self.member_groups = self.truss.member_groups
        
        # Add the truss nodes to the frame:
        n0 = len(self.nodes)        
        
        for node in self.truss.nodes:
            node.node_id += n0
            self.nodes.append(node)
                
        #for mem in self.truss.members.values():
        #self.add(mem)
        self.members.update(self.truss.members)
        
        """
        self.truss.top_chord[0].hinges[0] = True
        self.truss.top_chord[-1].hinges[1] = True
        
        if self.bottom_chord_to_column:
            self.truss.bottom_chord[0].hinges[0] = True
            self.truss.bottom_chord[-1].hinges[1] = True
        """
    def create_truss_to_column_joints(self,y_joint_node=False):
        """ Creates joints between columns and joints 
            input:
                y_joint_node .. True, if the Y-joint at the chord is given a new
                node.
        """
        
        for col, ndx, left in zip(self.columns,[0,-1],[True,False]):
            # Top chord joints            
            newJoint = Truss2ColumnJoint(col.nodes[-1],self.truss.top_chord[ndx],col,left)
            self.column_joints.append(newJoint)
            
            # Make hinge to the top chord member
            # NOTE: the index '1' works for both ends of the bottom chord, because
            # the element numbering of the left-hand-side top chord member starts from
            # the first K-joint in the top chord and goes to the column.
            # See: raami_plane_truss.py -> TopChord -> generate_elements
            #self.truss.top_chord[ndx].hinges[1] = True
                            
            if y_joint_node:
                # Create new FrameNode which also replaces the node
                # at the Y joint of the chord.
                for joint in newJoint.chord.joints:
                    if isinstance(joint,TubularYJoint):
                        newCoord = np.array(copy(joint.node.coords))
                        D = 0.5*newJoint.hc*abs(np.cos(newJoint.chord.angle)) + newJoint.gap
                        if left:                            
                            newCoord = newCoord + D*newJoint.chord.dir_vector
                        else:
                            newCoord = newCoord - D*newJoint.chord.dir_vector
                                                
                        newNode = FrameNode(newCoord)                    
                        self.add(newNode)
                        joint.node = newNode
                                    
            if self.bottom_chord_to_column:
                newJoint = Truss2ColumnJoint(col.nodes[1],self.truss.bottom_chord[ndx],col,left)
                self.column_joints.append(newJoint)
                
                #if left:
                #    self.truss.bottom_chord[ndx].hinges[0] = True
                #else:
                #    self.truss.bottom_chord[ndx].hinges[1] = True
                if y_joint_node:
                    # Create new FrameNode which also replaces the node
                    # at the Y joint of the chord.
                    for joint in newJoint.chord.joints:
                        if isinstance(joint,TubularYJoint):
                            newCoord = copy(joint.node.coords)
                            
                            # Determine the index of the node among the
                            # FrameNodes of the brace of the Y-joint
                            y = newCoord[1]
                            if joint.braces[0].nodes[0].y == y:
                                brace_node_nd = 0
                            else:
                                brace_node_nd = 1
                            
                            D = 0.5*newJoint.hc*abs(np.cos(newJoint.chord.angle)) + newJoint.gap
                            if left:
                                newCoord += D*newJoint.chord.dir_vector
                            else:
                                newCoord -= D*newJoint.chord.dir_vector
                            
                            #print(newCoord,joint.node.coords)
                            newNode = FrameNode(newCoord)       
                            self.add(newNode)
                            joint.node = newNode
                            joint.braces[0].nodes[brace_node_nd] = newNode
            
            
    
    def generate_supports(self,col_base="fixed"):
        """ Creates supports to the truss """
        
        if col_base == 'fixed':
            print(self.nodes[0],self.nodes[1])
            self.add(FixedSupport(self.nodes[0]))
            self.add(FixedSupport(self.nodes[1]))

        elif col_base == 'pinned':
            self.add(XYHingedSupport(self.nodes[0]))
            self.add(XYHingedSupport(self.nodes[1]))

    def generate_fem(self,truss_eccentricity=False,column_eccentricity=False):
        """ Creates the finite element model """
        if not self.fem_generated:
            self.fem_generated = True
        
        if self.bottom_chord_to_column:
            # In this case, the bottom chord is connected to the column
            # First, generate columns                       
            for col in self.columns:
                for node in col.nodes:
                    if self.dim == 2:
                        newNode = self.fem.add_node(node.x,node.y)
                    else:
                        newNode = self.fem.add_node(node.x,node.y,node.z)
                    
                    node.fem_node = newNode
            
            for member in self.columns:            
                member.generate_elements(self.fem)
                
            # Generate eccentricity elements, if any
            if column_eccentricity:
                for joint in self.column_joints:
                    if joint.left:
                        dx = 0.5*joint.hc
                    else:
                        dx = -0.5*joint.hc
                    
                    dy = 0.5*joint.hc*abs(np.tan(joint.chord.angle))
                    newNode = self.fem.add_node(joint.node.x+dx,joint.node.y+dy)
                                        
                    joint.fem_nodes['xc'] = newNode
                    
                    n1 = joint.node.fem_node
                    n2 = joint.fem_nodes['xc']
                    cs = IPE(500)
                                 
                    newElement = EBBeam(n1,n2,cs,cs.material)                    
                    
                    # Add hinge to the end of the member
                    #newElement.releases = [5]
                    
                    self.fem.add_element(newElement)
                    joint.fem_elements = newElement
            
            # Then, generate truss, including loads on trusses
            self.truss.generate_fem("ecc_elements",nodes_and_elements_only=False)
            
            self.truss.top_chord[0].fem_elements[-1].releases = [5]
            self.truss.top_chord[-1].fem_elements[-1].releases = [5]
            
            if self.bottom_chord_to_column:
                # Finally, connect the bottom chords to the eccentricity elements                
                n1 = self.column_joints[1].fem_nodes['xc']
                n2 = self.column_joints[1].chord.joints[0].fem_nodes['xc']                
                cs = self.column_joints[1].chord.cross_section
                newElement = EBBeam(n1,n2,cs,cs.material)                
                self.fem.add_element(newElement)
                self.column_joints[1].chord.fem_elements.insert(0,newElement)
                
                self.column_joints[1].chord.fem_elements[0].releases = [2]                
                
                n2 = self.column_joints[3].fem_nodes['xc']
                n1 = self.column_joints[3].chord.joints[1].fem_nodes['xc']
                cs = self.column_joints[3].chord.cross_section
                newElement = EBBeam(n1,n2,cs,cs.material)                
                self.fem.add_element(newElement)
                self.column_joints[3].chord.fem_elements.append(newElement)
            
                self.column_joints[3].chord.fem_elements[-1].releases = [5]
            
            #self.fem.draw()
        else:
            # In this case, the bottom chord is not connected to the column            
            # First, generate columns                       
            for col in self.columns:
                for node in col.nodes:
                    if self.dim == 2:
                        newNode = self.fem.add_node(node.x,node.y)
                    else:
                        newNode = self.fem.add_node(node.x,node.y,node.z)
                    
                    node.fem_node = newNode
            
            for member in self.columns:            
                member.generate_elements(self.fem)
            
            # Generate eccentricity elements, if any
            if column_eccentricity:
                for joint in self.column_joints:
                    if joint.left:
                        dx = 0.5*joint.hc
                    else:
                        dx = -0.5*joint.hc
                    
                    dy = 0.5*joint.hc*abs(np.tan(joint.chord.angle))
                    newNode = self.fem.add_node(joint.node.x+dx,joint.node.y+dy)
                                        
                    joint.fem_nodes['xc'] = newNode
                    
                    n1 = joint.node.fem_node
                    n2 = joint.fem_nodes['xc']
                    cs = IPE(500)
                                 
                    newElement = EBBeam(n1,n2,cs,cs.material)                    
                    
                    # Add hinge to the end of the member
                    #newElement.releases = [5]
                    
                    self.fem.add_element(newElement)
                    joint.fem_elements = newElement
            
            # Then, generate truss, including loads on trusses
            self.truss.generate_fem("ecc_elements",nodes_and_elements_only=False)
            
            #self.truss.top_chord[0].fem_elements[-1].releases = [5]
            #self.truss.top_chord[-1].fem_elements[-1].releases = [5]
            
            # Finally, connect the top chords to the eccentricity elements                
            n1 = self.column_joints[0].fem_nodes['xc']
            n2 = self.column_joints[0].chord.joints[0].fem_nodes['xc']                
            cs = self.column_joints[0].chord.cross_section
            newElement = EBBeam(n2,n1,cs,cs.material)                
            self.fem.add_element(newElement)
            self.column_joints[0].chord.fem_elements.insert(0,newElement)
            
            self.column_joints[0].chord.fem_elements[0].releases = [2]                
            
            n2 = self.column_joints[1].fem_nodes['xc']
            n1 = self.column_joints[1].chord.joints[1].fem_nodes['xc']
            cs = self.column_joints[1].chord.cross_section
            newElement = EBBeam(n2,n1,cs,cs.material)                
            self.fem.add_element(newElement)
            self.column_joints[1].chord.fem_elements.append(newElement)
        
            self.column_joints[1].chord.fem_elements[-1].releases = [5]
        
        #self.fem.draw()
        
        # Generate loads
        for load in self.loads.values():
            load.add_load(self.fem)
            # Creates new loadcase if one with same load_id does not exist
            lcase_ids = [lc.load for lc in self.fem.loadcases.values()]
            if load.load_id not in lcase_ids:
                self.fem.add_loadcase(supp_id=1,load_id=load.load_id)
        
        # Generate supports
        for supp in self.supports.values():
            supp.add_support(self.fem)

        # Set nodal degrees of freedom        
        self.fem.nodal_dofs()
        
        """
        # Generate finite element model of the truss        
        
        if truss_eccentricity:
            self.truss.generate_fem("en1993")
        else:
            self.truss.generate_fem("no_eccentricity")
        
        
        # Add fem nodes for the column bases
        for node in self.nodes[:2]:
            if self.dim == 2:
                newNode = self.fem.add_node(node.x,node.y)
            else:
                newNode = self.fem.add_node(node.x,node.y,node.z)
            
            node.fem_node = newNode
       
        
        # Generate finite elements for the columns        
        for member in self.columns:            
            member.generate_elements(self.fem)
        
        # Generate supports
        for supp in self.supports.values():
            supp.add_support(self.fem)

        # Set nodal degrees of freedom        
        self.fem.nodal_dofs()
        """
    
    def clear_fem(self):
        
        super().clear_fem()
        self.truss.clear_fem()
        self.truss.fem = self.fem
        
    
    def update_fem_joint_nodes(self):
        """ Updates nodal positions at joints. It is assumed that cross-sections
            have been changed along the way.
        """
        
        # Update column joint FEM-nodes. This means
        # 1) The node at the column face is moved according to the new column section
        # 2) 
        for joint in self.column_joints:
            if joint.left:
                dx = 0.5*joint.hc
                dg = joint.gap
            else:
                dx = -0.5*joint.hc
                dg = -joint.gap
                
            
            dgy = joint.gap*abs(np.tan(joint.chord.angle))
            dy = 0.5*joint.hc*abs(np.tan(joint.chord.angle))
            joint.fem_nodes['xc'].x = joint.node.x+dx
            joint.fem_nodes['xc'].y = joint.node.y+dy
            
            for mem in joint.node.members:
                # Update the location of the brace node close to the column
                if isinstance(mem,TrussBrace):
                    if isinstance(mem.top_joint,TubularYJoint):                        
                        mem.top_joint.node.x = joint.node.x+dx+dg
                        mem.top_joint.node.y = joint.node.y+dy+dgy
                    elif isinstance(mem.bottom_joint,TubularYJoint):
                        mem.bottom_joint.node.x = joint.node.x+dx+dg
                        mem.bottom_joint.node.y = joint.node.y+dy+dgy

    
        # Update truss joints FEM nodes
        self.truss.update_fem_joint_nodes()    
    
    def generate_uniform_load(self,part='truss',q=-20,member="top",load_id=LoadIDs['ULS'],ltype='live'):
        
        if part == 'truss':
            # Tässä generoidaan ristikolle kuormitustapaus, jonka load_id on "load_id"
            # Jos ristikolla on jo ko. load_id, lisätään tämä kuormitus ko. load_caseen.
            # Ristikolla on "load_cases" attribuutti, johon kuormitus lisätään.
            #
            # Nyt pitäisi sitten lisätä tämä ristikon kuormitus myös PortalTruss-oliolle.
            #
            # a) jos PortalTruss-luokalla on jo load_id:n omaava kuormitustapaus, lisätään
            #    kuorma tälle.
            #    - voidaan kopioida suoraan ristikolta load_id:n load_case myös kehälle.
            #      tässä voi olla se riski, että jos ristikolle halutaan useita kuormia samalla id:llä,
            #      voi tämä kopiointi epäonnistua. Ehkä dict:n update-metodi toimii tässä.
            #    - update toiminee, jos kaikki ristikon kuormat annetaan ennen pilareiden kuormia.
            #
            self.truss.generate_uniform_load(q,member,load_id,ltype)
            
            if not (load_id in self.load_ids):
                self.load_ids.append(load_id)
                self.add(self.truss.load_cases[load_id])
                
            """
            if not (this.load_id in self.load_ids):
                
                NewLoadCase = LoadCase(load_id=this.load_id,loads = [this])
                
                self.add(NewLoadCase)
                self.load_ids.append(this.load_id)
            else:
                # If a load case with this.load_id exists, add the new
                # load to that load case
                self.load_cases[this.load_id].add(this)
                
            if this.name in self.loads.keys():
                this.name += str(len(self.loads))
            self.loads[this.name] = this
            """
            
        elif part == 'column':
            if not (load_id in self.load_ids):
                self.load_ids.append(load_id)
            
            load_name = 'COL'
            if member == 'left':
                mem = self.columns[0]
                load_name += '_left'
            else:
                mem = self.columns[1]
                load_name += '_right'
            
            vals = [q,q]
            
            self.add(LineLoad(mem, vals, 'x', load_id, 1.0, ltype, load_name, "global"))
    
    def generate_piecewise_uniform_load(self,q,xzones,load_id=LoadIDs['ULS'],ltype='live'):
        """ Adds a piecewise uniform load. See the same method for SlopedTruss in 'raami_plane_truss.py' """
        self.truss.generate_piecewise_uniform_load(q,xzones,load_id,ltype)
        
        if not (load_id in self.load_ids):
            self.load_ids.append(load_id)
            self.add(self.truss.load_cases[load_id])
    
    def plot(self,print_text=True, show=True,
             loads=True, color=False, axes=None, save=False, mem_dim=False, geometry=False,
             saveopts={'filename':'PortalTruss','format':'svg','orientation':'landscape', 'papertype':'a3'}):
                
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        plt.rcParams['text.color'] = "red"
        plt.rcParams['font.size'] = 6.0
        self.truss.plot(geometry=geometry,print_text=print_text,show=show,
                        loads=loads,color=color,axes=ax,save=False,mem_dim=mem_dim)
        
        for col in self.columns:
            col.plot(print_text=print_text, c=color, axes=ax, mem_dim=mem_dim)
        
        for support in self.supports.values():
            node_coord = support.node.coords
            if support.dofs == [-1, -1, -1] or support.dofs == [1, 1, 1] or support.dofs == [0, 1, 2]:
                marker = 's'
            elif support.dofs == [1]:
                marker = '^'
            elif support.dofs == [0]:
                marker = '>'
            else:
                marker = 'D'
            ax.scatter(node_coord[0], node_coord[1], s=50, c='k',
                        marker=marker)
        

        # if self.truss:
        #    self.truss.plot(show=False, print_text=print_text, color=color)
        if loads:
            self.plot_loads()
            
        ax.axis('equal')
        if save:
            #plt.savefig('default.svg', format='svg')
            print("Saving figure.")
            plt.savefig(saveopts['filename'],format=saveopts['format'],\
                        orientation=saveopts['orientation'],papertype=saveopts['papertype'])
        if show:
            plt.axis('equal')
            plt.show()
    
    def optimize_members(self, prof_type="CURRENT",verb=False,limit_width=False,**kwargs):    
        """ Finds minimum weight profiles for members """
               
        top = {'material': 'S355', 'class': 2, 'utility': 1.0, 'bmin':10, 'bmax':1e5}
        bottom = {'material': 'S355', 'class': 2, 'utility': 1.0,'bmin':10, 'bmax':1e5}
        braces = {'material': 'S355', 'class': 2, 'utility_tens': 1.0, 'utility_comp':1.0, 'bmin':10, 'bmax':1e5}
        columns = {'material': 'S355', 'class': 2, 'utility': 1.0, 'bmin':10, 'bmax':1e5} 
        
        max_slenderness = 5.0
        
        for key, value in kwargs.items():
            if key == 'top':
                top.update(value)
            elif key == 'bottom':
                bottom.update(value)
            elif key == 'braces':
                braces.update(value)
            elif key == 'columns':
                columns.update(value)
            elif key == 'min_width':
                min_width = value
            elif key == 'max_slenderness':
                max_slenderness = value
        
        kmax = 10
        k = 1
                        
        while k < kmax:
            explored = []
            MEM_PROFILES_CHANGED = []
            self.structural_analysis('all','REM')
            if verb:
                print(f"Optimize profiles, iteration {k}")
            
            if limit_width:
                b_chord_min = self.truss.max_brace_width()
                #b_brace_max = min(self.top_chord_profile.b,self.bottom_chord_profile.b)
                b_brace_max = 1e5
            else:
                b_chord_min = 10
                b_brace_max = 1e5
            
            # Go through groups first
            for name, group in self.member_groups.items():
                if name == "top_chord":
                    material = top['material']
                    sec_class = top['class']
                    max_utility = top['utility']
                    b_min = max(top['bmin'],b_chord_min)
                elif name == 'bottom_chord':
                    material = bottom['material']
                    sec_class = bottom['class']
                    max_utility = bottom['utility']
                    b_min = max(bottom['bmin'],b_chord_min)
                elif name == 'columns':
                    material = columns['material']
                    sec_class = columns['class']
                    max_utility = columns['utility']
                    b_min = columns['bmin']
                    print(f"Column minimum width: {b_min:4.2f} mm")
                
                print('Optimize ' + name)
                
                group.optimum_design(prof_type,verb,material,sec_class,max_utility,b_min,
                                     max_slenderness=max_slenderness)
                
                for mem in group.members:
                    explored.append(mem)
                
            
            for member in self.members.values():
                #print(member)                
                if not member in explored:
                    if verb:
                        print(member)
                    explored.append(member)
                    if isinstance(member,TrussBrace):
                        material = braces['material']
                        sec_class = braces['class']
                        if list(member.NEd.values())[0] >= 0.0:
                            max_utility = braces['utility_tens']
                        else:
                            max_utility = braces['utility_comp']
                    else:
                        material = 'S355'
                        sec_class = 2
                        max_utility = 1.0
                        
                    MEM_PROFILES_CHANGED.append(member.optimum_design(prof_type,verb,material,sec_class,max_utility,
                                                                      bmax=b_brace_max,max_slenderness=max_slenderness))
                                        
                    if not member.symmetry_pair is None:
                        explored.append(member.symmetry_pair)
                    
                    #print(member.cross_section)
            
            # If none of the members changed profile, exit the loop
            if not any(MEM_PROFILES_CHANGED):
                break
            else:
                # Update nodal positions of FEM nodes at joints
                self.update_fem_joint_nodes()
                #self.fem.draw()
            k += 1
        
        # Update nodal positions of FEM nodes at joints
        self.update_fem_joint_nodes()
        
        if verb:
            print(f"Number of iterations: {k}.")
            
    
    def to_abaqus(self,target_dir='./',filename="Raami",partname='Truss',options=AbaqusOptions()):
        """ Exports truss to abaqus 
            The export method of 'raami' is mainly used. The main point here is to
            write the MPCs corresponding to eccentricity elements
        """
        
        # options
        #opts = {'x_monitor':0.5*self.span, 'n_monitored':2, 'mpc':[],'ecc_elements':[]}
        
        # Create element sets for top chord and bottom chord
        # top chord is halved
        L2 = 0.5*self.L
        options.elsets['Top_chord_left'] = []
        options.elsets['Top_chord_right'] = []
        #options.elsets['Bottom_chord'] = []
        options.elsets['Bottom_chord_left'] = []
        options.elsets['Bottom_chord_right'] = []
        
        options.elsets['Column_left'] = []
        options.elsets['Column_right'] = []
        
        
        options.load_elsets['Top_chord_load'] = []
        options.load_elsets['Column_left_load'] = []
        options.load_elsets['Column_right_load'] = []
        
        options.load_elsets['Wind_GF'] = []
        options.load_elsets['Wind_H'] = []
        options.load_elsets['Wind_I'] = []
        
        options.load_nsets['Column_left_top_node'] = []
        options.load_nsets['Column_right_top_node'] = []
        
        for mem_id, brace in self.truss.braces.items():
            options.included_members.append('B' + str(mem_id))
        
        for mem in self.truss.top_chord:
            for el in mem.fem_elements:
                if all([n.x <= L2 for n in el.nodes]):
                    options.elsets['Top_chord_left'].append(el)
                else:
                    options.elsets['Top_chord_right'].append(el)
        
        for mem in self.truss.bottom_chord:
            for el in mem.fem_elements:
                #options.elsets['Bottom_chord'].append(el)
                if all([n.x <= L2 for n in el.nodes]):
                    options.elsets['Bottom_chord_left'].append(el)
                else:
                    options.elsets['Bottom_chord_right'].append(el)
        
        for mem in self.columns:
            for el in mem.fem_elements:
                #options.elsets['Bottom_chord'].append(el)
                if mem.nodes[0].x < L2:
                    options.elsets['Column_left'].append(el)
                else:
                    options.elsets['Column_right'].append(el)
        
        for joint in self.truss.joints.values():
            if isinstance(joint,TubularKGapJoint):
                if not(joint.fem_elements['ecc'] is None):
                    # This applies if the model 'en1993' was used
                    pass
                elif not(joint.fem_elements['left_ecc'] is None):
                    # Determine which node of the left eccentricity element
                    # is in the brace member. That is the slave node
                    n = joint.fem_elements['left_ecc'].nodes[0]
                    if n in joint.left_brace.fem_nodes:
                        options.mpc.append([n.nid+1,joint.fem_elements['left_ecc'].nodes[1].nid+1])
                    else:
                        options.mpc.append([joint.fem_elements['left_ecc'].nodes[1].nid+1,n.nid+1])
                    
                    n = joint.fem_elements['right_ecc'].nodes[0]
                    if n in joint.right_brace.fem_nodes:
                        options.mpc.append([n.nid+1,joint.fem_elements['right_ecc'].nodes[1].nid+1])
                    else:
                        options.mpc.append([joint.fem_elements['right_ecc'].nodes[1].nid+1,n.nid+1])
                    
                    options.ecc_elements.append(joint.fem_elements['left_ecc'])
                    options.ecc_elements.append(joint.fem_elements['right_ecc'])
                    
                    if joint.chord == "top":
                        if joint.ridge:
                            options.top_gap_elements += joint.fem_elements['gap']
                        else:
                            options.top_gap_elements.append(joint.fem_elements['gap'])
                    elif joint.chord == 'bottom':
                        options.bottom_gap_elements.append(joint.fem_elements['gap'])
        
        for el in options.top_gap_elements:
            if all([n.x <= L2 for n in el.nodes]):
                options.elsets['Top_chord_left'].append(el)
            else:
                options.elsets['Top_chord_right'].append(el)
        
        for el in options.bottom_gap_elements:
            #options.elsets['Bottom_chord'].append(el)
            if all([n.x <= L2 for n in el.nodes]):
                options.elsets['Bottom_chord_left'].append(el)
            else:
                options.elsets['Bottom_chord_right'].append(el)
        
        
            
        options.load_elsets['Top_chord_load'] = options.elsets['Top_chord_left'] + \
            options.elsets['Top_chord_right'] 
        options.load_elsets['Column_left_load'] = options.elsets['Column_left']
        options.load_elsets['Column_right_load'] = options.elsets['Column_right']
        
        # Define node sets for point loads at column ends
        if self.bottom_chord_to_column:
            options.load_nsets['Column_left_top_node'].append(self.columns[0].members[1].fem_nodes[-1])
            options.load_nsets['Column_right_top_node'].append(self.columns[1].members[1].fem_nodes[-1])
        else:
            options.load_nsets['Column_left_top_node'].append(self.columns[0].fem_nodes[-1])
            options.load_nsets['Column_right_top_node'].append(self.columns[1].fem_nodes[-1])
        
        # Determine elsets for wind zones F/G, H, and I
        e = 2*self.ridge_height()
        
        xI = 0.5*e
        xG = 0.1*e
        
        for el in options.load_elsets['Top_chord_load']:            
            x = [node.x for node in el.nodes]
            x.sort()
            
            if x[0] >= xI:
                options.load_elsets['Wind_I'].append(el)
            elif x[0] < xG:
                options.load_elsets['Wind_GF'].append(el)
            else:
                options.load_elsets['Wind_H'].append(el)
        
        
        # Add MPCs for eccentricities of truss-to-column connections        
        for joint in self.column_joints:
            # By default, there is a single element in a truss-to-column joint
            if not (joint.fem_elements is None):                
                el = joint.fem_elements
                # The indexing works, because the first FEM node of the
                # connection is always on the column center line
                options.mpc.append([el.nodes[1].nid+1,el.nodes[0].nid+1])
                options.ecc_elements.append(el)
        
        
        
        # Loads
        # Implementation: each load case is 
        # Vaihtoehdot:
        #for load_id, loads in self.load_cases.items():
            
        
        super().to_abaqus(target_dir,filename,partname,options)
            
class Truss2ColumnJoint:
    """ Class for defining joints between chord and column """
    
    def __init__(self,node,chord,column,left=True):
        
        self.gap = 30
        self.node = node
        self.chord = chord
        self.column = column
        self.left = left
        
        # 'col' .. node at the center line of the column
        # 'xc' .. node at the surface of the column face
        self.fem_nodes = {'col': None, 'xc': None}
        self.fem_elements = None
        
        try:
            self.chord.joints.append(self)
        except:
            pass
                
    
    @property
    def hc(self):
        """ Height of the column profile """
        try:
            hc = self.column.cross_section.h
        except:
            hc = self.column.cross_section.H
        
        return hc
 
        
if __name__ == "__main__":
    
    from timeit import default_timer as timer
    
    p = PortalTruss()
    
    p.create_truss(topology='K',ndiv=3,h2=2000,h1=1500,dx=000,first_diagonal_up=True)
    #p.truss.top_chord[0].add_hinge(0)
    #p.truss.top_chord[-1].add_hinge(1)
    #if p.bottom_chord_to_column:
    #    p.truss.bottom_chord[0].add_hinge(0)
    #    p.truss.bottom_chord[-1].add_hinge(1)
        
    p.generate_uniform_load('truss',q=-25,ltype='snow')
    
    #p.truss.generate_supports()
    p.create_columns(profile=SHS(250,8),hinges=[False,False])
    p.create_truss_to_column_joints(y_joint_node=True)
    p.generate_supports()
    p.symmetry()
    
    
    p.generate_uniform_load('column',q=9,member='left',load_id=LoadIDs['ULS'],ltype='wind')
    p.generate_uniform_load('column',q=5,member='right',load_id=LoadIDs['ULS'],ltype='wind')
    

    p.generate_uniform_load('truss',q=-18,ltype='snow',load_id=10)
    p.generate_uniform_load('column',q=9,member='left',load_id=10,ltype='wind')
    p.generate_uniform_load('column',q=5,member='right',load_id=10,ltype='wind')
    
    
    #p.generate_uniform_load('column',q=5,member='left')    
    
    #start = timer()
    print(p.weight())
    #end = timer()
    #print(end-start)
    #p.weight_detailed(True)
    #Wtop, Wbottom, Wbraces = p.truss.weight_detailed(True)
    p.generate_fem(truss_eccentricity=True,column_eccentricity=True)
        
    #print(p.weight())
    #p.structural_analysis(load_id=p.load_ids[0],support_method="REM")
    
    p.structural_analysis(load_id='all',support_method="REM")
    
    
    #p.optimize_members(verb=True)
    #p.bmd(scale=20,load_id=p.load_ids[0],loads=False)
    #p.fem.draw()
    #p.plot()
    
    opts = AbaqusOptions(x_monitor = 0.5*p.L, n_monitored = 2)
    p.to_abaqus(target_dir='C:/Users/kmela/Data/',filename='K-portaali',partname="K-portaali",options=opts)
        


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import scipy as sp
import time
from copy import deepcopy


from abc import ABCMeta, abstractclassmethod
from scipy.sparse.linalg import eigsh


class FrameFEM:
    """ Class for frame analysis by finite element method

        Written by Kristo Mela

        Variables:
        -------------------
        :ivar nodes: FEMNodes
        :ivar nodal_coords: nodal coordinates [x, y]
        :ivar elements: Element -instances
        :ivar sections: Section -instances
        :ivar materials: Material -instances
        :ivar supports: Support -instances
        :ivar loads: Load -instances
        :ivar loadcases: LoadCase -instances
        :ivar dofs: number of degrees of freedom

        :vartype nodes: list
        :vartype nodal_coords: list
        :vartype elements: list
        :vartype sections: list
        :vartype materials: list
        :vartype supports: list
        :vartype loads: list
        :vartype loadcases: dict
        :vartype dofs: int

    """

    def __init__(self):

        """ List containing all FEMNode -instances """
        self.nodes = []
        """ List of nodal coordinates """
        self.nodal_coords = []
        """ List of elements """
        self.elements = []
        """ List of sections """
        self.sections = []
        """ List of materials """
        self.materials = []
        """ List of supports """
        self.supports = []
        """ List of loads """
        self.loads = []
        """ Dict of load cases
            The key is the load id of the corresponding load case
            Value is a LoadCase object
        """
        self.loadcases = {}
        """ Number of degrees of freedom """
        self.dofs = 0
        """ Dimension of the problem (either 2 or 3) 
            Problem is 3D if any node as three coordinates
        """
        self.dim = 2
        """ Supported degrees of freedom """
        self.__supp_dofs = []
        self.supp_dofs = []
        
        self.load_factors = []
        self.buckling_modes = []
        
        
        self.u = []
        self.u2 = []
        
    def __repr__(self):
        
        s = f"Frame {self.dim}D model: {self.nels()} elements, {self.nnodes()} nodes"
        
        return s
        
    def nels(self):
        """ Number of elements

        Returns:
        ---------
        :return: number of elements
        :rtype: int
        """
        return len(self.elements)

    def nnodes(self):
        """ Number of nodes

        Returns:
        --------
        :return: number of nodes
        :rtype: int

        """
        return len(self.nodes)

    def nsupp(self):
        """ Number of supported nodes

        Returns:
        --------
        :return: number of supported nodes
        :rtype: int

        """
        return len(self.supports)

    def nload(self):
        """ Number of loads

        Returns:
        --------
        :return: number of loads
        :rtype: int
        """
        return len(self.loads)

    def nloadcases(self):
        """ Number of load cases

        Returns:
        --------
        :return: number of load cases
        :rtype: int
        """
        return len(self.loadcases)
    
    @property
    def supp_dofs(self):
        """ Get supported degrees of freedom """
        
        if len(self.__supp_dofs) == 0:
            self.supported_dofs()
        
        return self.__supp_dofs
    
    @supp_dofs.setter
    def supp_dofs(self,val):
        """ Set supported degrees of freedom """        
        
        self.__supp_dofs = val
        
    def add_node(self, x, y, z=None):
        """ Adds node to fem model

        Parameters:
        -----------
        :param x: x-coordinate for node
        :param y: y-coordinate for node

        :type x: float
        :type y: float
        """

        newNode = FEMNode(len(self.nodes), x, y, z)
        if z is None:
            """ if z is 'None', there is no z-coordinate, and the problem
                is 2D. Otherwise, the value of z-coordinate is given, and the problem
                is 3D.
            """

            self.nodal_coords.append([x, y])            
            for load_id in self.loadcases.keys():   
                newNode.u[load_id] = np.zeros(len(newNode.dofs))
                #newNode.u[load_id] = [0.0, 0.0, 0.0]
        else:
            self.dim = 3
            self.nodal_coords.append([x, y, z])
            #for i in range(self.nloadcases()):
            #    newNode.u = np.vstack((newNode.u, [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]))

        self.nodes.append(newNode)
        return newNode

    @property
    def nfree_dofs(self):
        """ number of free degrees of freedom """
        if len(self.supp_dofs) == 0:
            self.supported_dofs()
        
        nsupp_dofs = len(self.supp_dofs)
        
        return self.dofs-nsupp_dofs
    
    def element_free_dofs(self,el):
        """ Returns free dofs of an element """

        if isinstance(el,int):
            el = self.elements[el]
        
        el_dofs = el.global_dofs()
        supp_dofs = self.supp_dofs

        return [a for a in el_dofs if not a in supp_dofs]
        
        
    def add_material(self, E, nu, density, G=None):
        """ Adds new material

        Parameters:
        -----------
        :param E: Young's modulus [MPa]
        :param nu: Poisson ratio
        :param density: density [kg / mm^3]

        :type E: float
        :type nu: float
        :type density: float
        """
        newMat = Material(E, density, nu=nu, G=G)
        self.materials.append(newMat)
        return newMat

    def add_section(self, newSection):
        """ Add new cross-section

        Parameters:
        -----------
        :param newSection: section to be added to fem-model

        :type newSection: Section
        """
        self.sections.append(newSection)

    def add_element(self, new_elem):
        """ Add new element

        Parameters:
        -----------
        :param new_elem: element to be added

        :type new_elem: Element

        """
        
        for lcase in self.loadcases.keys():
            new_elem.add_loadcase(lcase)
        
        self.elements.append(new_elem)
        

    def add_support(self, sid, nid, dof, val=0):
        """ Add support condition

        Parameters:
        -----------
        :param sid: support id
        :param nid: index of supported node
        :param dof: degrees of freedom
        :param val: value for support displacement

        :type sid: int
        :type nid: int
        :type dof: list
        :type val: float

        """
        new_support = Support(sid, self.nodes[nid], dof, val)
        self.supports.append(new_support)

    def add_load(self, load):
        """ Add load

        Parameters:
        ------------
        :param load: load to be added

        :type load: Load
        """
        self.loads.append(load)

    def add_loadcase(self, supp_id=1, load_id=2):
        """ Adds new load case, if load id is same replaces old one

        Parameters:
        -----------
        :param supp_id: support id
        :param load_id: load id

        :type supp_id: int
        :type load_id: int
        """
        #print('Create Load case.')
        new_case = LoadCase(supp_id, load_id)
        self.loadcases[load_id] = new_case

        """ Add displacements to nodes for the new load case """
        ncases = self.nloadcases()
        for node in self.nodes:
            node.u[load_id] = np.zeros(len(node.dofs))
            node.u2[load_id] = np.zeros(len(node.dofs))
            """
            if ncases == 1:
                print(node.u[0])
                node.u[0] = np.zeros(len(node.dofs))
            else:
                node.u.append(np.zeros(len(node.dofs)))
            """
            """    
            if len(node.u) > 0:
                # node.u = [0.0, 0.0, 0.0]
                # node.u = np.vstack((node.u, [0.0, 0.0, 0.0]))
                node.u = np.vstack(node.u,[0 for i in range(len(node.dofs))])
            else:
                node.u = [0 for i in range(len(node.dofs))]
                # node.u = np.vstack((node.u, [0.0, 0.0, 0.0]))
            """        
        for el in self.elements:
            el.add_loadcase(load_id)
            
    def add_release(self,eid,dofs):
        """ Adds releases to an element
            input: eid .. element number
                   dofs .. list of degrees of freedom for the given element
                           for which the corresponding internal forces
                           are released (set to zero)
        """
        print("vapautus")
        self.elements[eid].releases = dofs
    
    def nodal_dofs(self):
        """ Assign nodal degrees of freedom
        """

        """ HERE THE ELEMENTS CAN BE USED TO SET SOME OF THE 
            DOFS TO ZERO. FOR EXAMPLE, TRUSS ELEMENTS DO NOT HAVE
            ROTATIONAL DEGREES OF FREEDOM, AND IF SOME NODES
            ARE CONNECTED TO ONLY TRUSS ELEMENTS, THOSE NODES WILL NOT
            HAVE A ROTATION.
        """
        self.dofs = 0
        
        for elem in self.elements:
            elem.init_dofs()

        """ Make running numbering of dofs.
            If the dof is marked as 1, then its dof is set to 0.
        """
        for n in self.nodes:
            for i in range(len(n.dofs)):
                #if n.dofs[i] > 0:
                """if not n.dofs[i]:
                    print('Not dof')
                    n.dofs[i] = -1
                else:
                """
                if n.dofs[i] >= 0:
                    #print('Dof')
                    n.dofs[i] = self.dofs
                    self.dofs += 1
            #print(n.dofs)



    def element_node_indices(self, e):
        """ Get node indices of an element

        Parameters:
        -----------
        :param e: element index

        :type e: int

        Returns:
        --------
        :return: list of element node indices
        :rtype: list
        """
        ndx = [self.nodes.index(n) for n in self.elements[e].nodes]
        # ndx = []
        # for n in self.elements[e].nodes:
        #    ndx.append(self.nodes.index(n))

        return ndx

    def global_free_dofs(self):
        """ List of free degrees of freedom """
        
        return list(set(np.arange(self.dofs))-set(self.supp_dofs))

    def global_stiffness_matrix(self):
        """ Constructs the global stiffness matrix

        Returns:
        --------
        :return: global stiffness matrix
        :rtype: np.array
        """

        """ initialize zero stiffness matrix 
            NOTE: for larger structures, K should probably be
            a sparse matrix
        """
        K = np.zeros((self.dofs, self.dofs))

        for elem in self.elements:
            ke = elem.stiffness_matrix()
            # get element degrees of freedom
            # change the list to numpy array
            ve = np.array(elem.global_dofs())
            
            # find non-zero dofs
            nz = ve >= 0
            q = ve[nz].astype(int)
            """ extract the submatrix of the element
                and add it to the global stiffness matrix
            """
            # a = np.arange(len(q))
            # row = np.vstack(a[nz])
            # col = a[nz]
            # K[np.vstack(q), q] += ke[row, col]
            #print(ke[np.ix_(nz, nz)])
            #print(ve,q)
            K[np.ix_(q, q)] += ke[np.ix_(nz, nz)]

        return K

    def global_geometric_stiffness_matrix(self,lcase=0):
        """ Constructs the global geometric stiffness matrix

        Returns:
        --------
        :return: global geometric stiffness matrix
        :rtype: np.array
        """
        dofs = self.dofs

        """ initialize zero geometric stiffness matrix 
            NOTE: for larger structures, KG should probably be
            a sparse matrix
        """
        KG = np.zeros((dofs, dofs))

        for elem in self.elements:
            kg = elem.geometric_stiffness_matrix(lcase)
            # get element degrees of freedom
            # change the list to numpy array
            ve = np.array(elem.global_dofs())
            # find non-zero dofs
            nz = ve >= 0
            q = ve[nz]

            """ extract the submatrix of the element
                and add it to the global stiffness matrix
            """
            KG[np.ix_(q, q)] += kg[np.ix_(nz, nz)]

        return KG

    def global_second_order_stiffness_matrix(self,lcase=0):
        """ Global stiffness matrix for second order analysis
        
        Returns:
        --------
        :return: global second order stiffness matrix
        :rtype: np.array
        """
        dofs = self.dofs

        """ initialize zero stiffness matrix 
            NOTE: for larger structures, KG should probably be
            a sparse matrix
        """
        K = np.zeros((dofs, dofs))

        for elem in self.elements:
            k = elem.second_order_stiffness_matrix(lcase)
            # get element degrees of freedom
            # change the list to numpy array
            ve = np.array(elem.global_dofs())
            # find non-zero dofs
            nz = ve >= 0
            q = ve[nz]

            """ extract the submatrix of the element
                and add it to the global stiffness matrix
            """
            K[np.ix_(q, q)] += k[np.ix_(nz, nz)]

        return K
        

    def global_load_vector(self, sid, order=1):
        """ Constructs global load vector

        Parameters:
        -----------
        :param sid: support id
        :type sid: int

        Returns:
        --------
        :return: global load vector
        :rtype: np.array
        """

        """ Initialize load vector """
        global_load = np.zeros(self.dofs)

        for load in self.loads:
            """ Get the loads to be added to the vector 
                and the global degrees of freedom of these loads
            """
            if load.sid == sid:                
                #print('Calculate nodal forces')
                v, vdofs = load.load_and_dofs(order)
                #print(v,vdofs)
                global_load[vdofs.astype(int)] += v

        # print("FrameFEM  ", '\n', global_load)
        return global_load
    
    def supported_dofs(self):
        """ Determine supported dofs. 
        
            Assumption: there is only one support case, so no need to
            distinguish between different support IDs.
        """
        supp_dofs = []
        for supp in self.supports:
            supp_dofs += list(supp.node.dofs[supp.dof])      
        
        self.supp_dofs = supp_dofs

    def linear_statics(self, lcase=0, support_method='ZERO'):
        """ Perform linear elastic static analysis for a given load case

        Parameters:
        -----------
        :param lcase: index for load case
        :param support_method: how supported DOFS are treated
                              'ZERO' .. columns and rows corresponding to
                              supported DOFS are zeroed and a 1 is put
                              to the diagonal
                              
                              'REM' .. columns and rows corresponding to the
                              supported DOFS are removed.

        :type lcase: int
        
        """
        if not self.dofs:
            self.nodal_dofs()

        #start = time.time()
        K = self.global_stiffness_matrix()
        
        if not isinstance(lcase,list):
            lcase = [lcase]
        
        #print(lcase)
        #print(self.loadcases)
        
        #load_id = self.loadcases[lcase].load        
        
        
        load_ids = []
        # Loads is a matrix whose columns are load vectors in
        # the different load cases to be analysed
        loads = []
        for lc in lcase:
            load_ids.append(self.loadcases[lc].load)
            loads.append(self.global_load_vector(load_ids[-1]))
        
        #p = self.global_load_vector(load_id)
        
       
        #print(p)
        """ Take supports into account
            The method is based on Filippa's Lecture Notes (3.5.2)
            
            The idea is that the rows and columns of the global stiffness
            matrix corresponding to the supported degrees of freedom are
            zeroed, and 1 is placed on the diagonal. Similarly, the
            elements of the load vector corresponding to the supported
            DOFs are zeroed.
        """
        
        # If there are several load cases, it is assumed that all of them
        # have the same support conditions.
        supp_id = self.loadcases[lcase[0]].support
        
        rem_dofs = []
        """ Vector of global unsupported dofs """
        self.glob_dofs = np.arange(self.dofs)
        
        
        for supp in self.supports:
            if supp.sid == supp_id:                
                node_dofs = supp.node.dofs
                for dof in supp.dof:
                    # Get the global degree of freedom of the supported
                    # node's degree of freedom.
                    i = node_dofs[dof]
                    #  print(i)
                    if i > -1:
                        if support_method == 'ZERO':
                            K[i,:] = 0
                            K[:,i] = 0
                            K[i,i] = 1
                            for p in loads:
                                p[i] = 0
                        elif support_method == 'REM':
                            rem_dofs.append(i)
        
        self.supp_dofs = rem_dofs
        
        
        if support_method == 'REM':
            rem_dofs = np.array(rem_dofs,dtype=int)
            
            K = np.delete(K,rem_dofs,0)
            K = np.delete(K,rem_dofs,1)
            LOADS = [np.delete(p,rem_dofs) for p in loads]
            #for p in loads:
            #    p = np.delete(p,rem_dofs)            
            
            self.glob_dofs = np.delete(self.glob_dofs,rem_dofs)
            
            #print(rem_dofs)
            #print(K)
            #print(p)
            self.K = K
            self.p = np.column_stack(LOADS)
            #print(glob_dofs)
                        
        """ Solve global displacement vector """

        #Kc, low = sp.linalg.cho_factor(K)
        #uc = sp.linalg.cho_solve((Kc, low), p)

        # print(f'K shape {K.shape}')
        # print(f'K {K}')
        # print(f'p {p}')
        u = np.linalg.solve(K, self.p)
        self.u = u
        #print(u)

        """ Substitute obtained displacements to nodes """
        if support_method == 'ZERO':
            """ Distribute displacements to nodes """
            for node in self.nodes:
                # Get nodal dofs
                dofs = np.array(node.dofs)
    
                # Find free dofs
                free_dofs = dofs >= 0
                # Substitute free dofs from global displacement vector                        
                for i, free_dof in enumerate(free_dofs):
                    if free_dof:
                        node.u[lcase][i] = u[dofs[i]]

            """
            This substition did not work !!!
            node.u[free_dofs] = u[dofs[free_dofs]]
            """
        elif support_method == 'REM':
            for i, lc in enumerate(lcase):                
                for ui, d in zip(u[:,i],self.glob_dofs):                
                    for node in self.nodes:
                        try:
                            #node.u[lcase][node.dofs.index(d)] = np.array(ui)
                            node.u[lc][np.where(node.dofs==d)] = np.array(ui)
                        except ValueError:
                            pass
                                    
        """ Calculate element internal forces """
        for i, el in enumerate(self.elements):     
            for lc in lcase:
                el.internal_forces(lcase=lc)
        #end = time.time()
        # print("FRAMEFEM TIME: ", end - start)
        return u, K
    
    def second_order_analysis(self, lcase=0, support_method="ZERO"):
        """ Performs second order elastic analysis """
        
        # Linear statics is first carried out to calculate the axial force
        # which is needed in second order analysis
        u, K = self.linear_statics(lcase,support_method)
        
        # Evaluate second order stiffness matrix
        KII = self.global_second_order_stiffness_matrix(lcase)
        
        # Evaluate second order load vector
        pII = self.global_load_vector(lcase,order=2)
                
        supp_id = self.loadcases[lcase].support
        
        # Supports        
        if support_method == 'ZERO':
            for i in self.supp_dofs:
                KII[i,:] = 0
                KII[i,:] = 0
                KII[:,i] = 0
                KII[i,i] = 1
                pII[i] = 0
        else:        
            KII = np.delete(KII,self.supp_dofs,0)
            KII = np.delete(KII,self.supp_dofs,1)
            pII = np.delete(pII,self.supp_dofs)
            
        self.KII = KII
        self.pII = pII
        
        print(pII)
            
        u2 = np.linalg.solve(KII, pII)
        
        #print(self.u)
        #print(u2)
        self.u2 = u2
        
        """ Substitute obtained displacements to nodes """
        if support_method == 'ZERO':
            """ Distribute displacements to nodes """
            for node in self.nodes:
                # Get nodal dofs
                dofs = np.array(node.dofs)
    
                # Find free dofs
                free_dofs = dofs >= 0
                # Substitute free dofs from global displacement vector                        
                for i, free_dof in enumerate(free_dofs):
                    if free_dof:
                        node.u2[lcase][i] = u2[dofs[i]]

            """
            This substition did not work !!!
            node.u[free_dofs] = u[dofs[free_dofs]]
            """
        elif support_method == 'REM':
            for ui, d in zip(u2,self.glob_dofs):                
                for node in self.nodes:
                    try:
                        #node.u[lcase][node.dofs.index(d)] = np.array(ui)
                        node.u2[lcase][np.where(node.dofs==d)] = np.array(ui)
                    except ValueError:
                        pass

        """ Calculate element internal forces """
        for i, el in enumerate(self.elements):            
            el.second_order_internal_forces(lcase=lcase)

    def second_order_elastic(self,lcase=0,ninc=10):
        """ Second order iterative elastic analysis """
                
        def tangent_stiffness(fem,lcase):
            """ Evaluate tangent stiffness matrix """
            u, Ke = fem.linear_statics(lcase,support_method="REM")
            Kg = fem.global_geometric_stiffness_matrix(lcase)
            Kg = np.delete(Kg,fem.supp_dofs,0)
            Kg = np.delete(Kg,fem.supp_dofs,1)
            
            return Ke+Kg
        
        def set_load_increment(fem,lcase,f):
            for load in fem.loads:
                if load.sid == lcase:
                    load.f = f
                    
        def update_nodal_coordinates(fem,lcase):
            for node in fem.nodes:
                node.X[:2] += node.u[lcase][:2]
        
        p = self.global_load_vector(lcase)
        p = np.delete(p,self.supp_dofs)
        d = 1/ninc
        dp = d*p
        du = np.zeros((ninc,self.nfree_dofs))
        print(du.shape)
        for i in range(ninc):            
            set_load_increment(self,lcase,i*d)
            Kt = tangent_stiffness(self,lcase)            
            du[i,:] = np.linalg.solve(Kt, dp)
            update_nodal_coordinates(self,lcase)
        
        print(du)
        
            

    def linear_buckling(self, lcase=0, k=4):
        """ Perform linear buckling analysis for a given load case

        Parameters:
        ------------
        :param lcase: load case id
        :param k: number of different buckling modes desired

        :type lcase: int
        :type k: int

        Returns:
        ---------
        :return: Array of k alpha_cr values and an vector representing buckling displacements

        :rtype: (np.array, np.array)
        
        TODO!!!
            the global geometric stiffness matrix does not include any fixing of
            supported dofs now.
            
            This is because the logic for fixing supported dofs in linear statics was
            changed from row and column elimination to setting elements to zero or one.
            
            Solution:
                when using linear buckling analysis, always use "REM" as method of
                support handling. Then, remove the rows and columns of the global
                geometric stiffness matrix as well.
        """

        """ Solve internal forces of members 
            and return stiffness matrix
        """
        u, K = self.linear_statics(lcase,support_method="REM")
        KG = self.global_geometric_stiffness_matrix(lcase)
        
        """
        for supp in self.supports:
            if supp.sid == supp_id:                
                node_dofs = supp.node.dofs
                for dof in supp.dof:
                    # Get the global degree of freedom of the supported
                    # node's degree of freedom.
                    i = node_dofs[dof]
                    #  print(i)
                    if i > -1:
                        if support_method == 'ZERO':
                            KG[i,:] = 0
                            KG[:,i] = 0
                            KG[i,i] = 1                            
                        elif support_method == 'REM':
                            rem_dofs.append(i)
        
        if support_method == 'REM':
            KG = np.delete(K,rem_dofs,0)
            KG = np.delete(K,rem_dofs,1)            
            glob_dofs = np.delete(glob_dofs,rem_dofs)
        """
        KG = np.delete(KG,self.supp_dofs,0)
        KG = np.delete(KG,self.supp_dofs,1) 
        
        
        # v0 -- initial guess is set to np.ones -matrix,
        # this makes sure that w and v values doesn't change
        # after consecutive runs, default v0 is matrix with random values
        (w, buckling_modes) = eigsh(K, k=k, M=-KG, sigma=1.0, which='LA', mode='buckling')
        #w = sp.linalg.eigvals(K,b=-KG)
        #(w,buckling_modes) = sp.linalg.eig(K,b=-KG)
        #(w,buckling_modes) = sp.sparse.linalg.eigs(K,k=6,M=-KG,which='SM',sigma=0)

        self.load_factors = w
        self.buckling_modes = buckling_modes

        """ Distribute buckling displacements to nodes """
        
        """
        for node in self.nodes:
               
            # Get nodal dofs
            
            dofs = np.array(node.dofs)
            
            # Find free dofs
            free_dofs = dofs >= 0
    
            # Substitute free dofs from global displacement vector
                                        
            for i in range(len(free_dofs)):
                if free_dofs[i]:
                    node.v[i] = v[dofs[i]]
        """
        
        for node in self.nodes:
            node.v = np.zeros((buckling_modes.shape[1],len(node.v)))
        
        
        glob_dofs = self.global_free_dofs()
        
        for i in range(buckling_modes.shape[1]):
            v = buckling_modes[:,i]
            for vi, d in zip(v,glob_dofs):                
                for node in self.nodes:
                    try:
                        node.v[i][list(node.dofs).index(d)] = vi
                    except ValueError:
                        pass
        
        return w, buckling_modes, KG

    def statics_matrix(self):
        """ Creates the statix matrix. This applicable for instances that
            only have Rod elements (extensible to EBBeam elements as well).
            
            Primary use of this method is for optimization of trusses (and frames)
            using the SAND approach
        """
        
        B = np.zeros([self.dofs,self.nels()])
        
        e = np.array([-1,1])
        for i, el in enumerate(self.elements):
            b = el.transformation_matrix().transpose().dot(e)
            d = el.global_dofs()
            B[np.ix_(d),i] = b
                        
        
        return np.delete(B,self.supp_dofs,0)
        
        #return B
        

    #def draw(self, deformed=False, buckling_mode=None, scale=1.0, axes=None):
    def draw(self, axes=None,**kwargs):
        #deformed=False, buckling_mode=None, scale=1.0, axes=None):
        """  Plots elements and nodes using matplotlib pyplot
        """

        #fig = plt.figure()
        if self.dim == 3:
            #ax = plt.axes(projection='3d')
            fig = plt.figure()
            ax = Axes3D(fig)
        else:
            if axes is None:
                fig, ax = plt.subplots(1)
            else:
                ax = axes

        """ draw nodes """
        
        for i, node in enumerate(self.nodes):
            node.plot(print_text=str(i),axes=ax)
        
        """ draw elements """
        if 'deformed' in kwargs:
            deformed = True
        else:
            deformed = False
        
        if 'buckling_mode' in kwargs:
            buckling_mode = True
            buck_mode = kwargs['buckling_mode']
        else:
            buckling_mode = False
        
        if 'scale' in kwargs:
            scale = kwargs['scale']
        else:
            scale = 1.0
        
        """ Draw elements """
        for i, el in enumerate(self.elements):
            """ Draw deformed shape is requested.
                the value of 'deformed' is the load case to be plotted
            """
            if 'deformed' in kwargs:
                lcase = kwargs['deformed']
                
                el.plot(print_text=str(i),axes=ax,deformed=lcase,scale=scale)
            else:
                el.plot(print_text=str(i),axes=ax)
                
        el_col = 'k'
        if deformed or buckling_mode is not None:
            lstyle = '--' #+ el_col
        else:
            lstyle = '-' + el_col
        
        # Harmaa: color = (0.6,0.6,0.6)
        if deformed:
            """
            for n in self.nodes:
                if self.dim == 2:
                    ax.plot(n.coord[0]+scale*n.u[0], n.coord[1]+scale*n.u[1], 'bo')
                else:
                    ax.scatter3D(n.coord[0], n.coord[1], n.coord[2],'ro')
            """
            lstyle = '-k'
            for el in self.elements:
                X = np.zeros((2,2))
                for i, n in enumerate(el.nodes):
                    X[i,:] = (n.coord + scale * np.array(n.u[:2]))
                    
                if self.dim == 2:
                    ax.plot(X[:, 0], X[:, 1], lstyle)                  
                else:
                    ax.plot3D(X[:, 0], X[:, 1], X[:,2], lstyle)
                
        
        if buckling_mode:
            lstyle = '-k'
            for el in self.elements:
                #print(el.nodes)
                X = np.zeros((2,2))
                for i, n in enumerate(el.nodes):
                    #print(i, buckling_mode, n.v[buck_mode][:2])
                    X[i,:] = (n.coord + scale * n.v[buck_mode][:2])
                    
                if self.dim == 2:
                    ax.plot(X[:, 0], X[:, 1], lstyle)                  
                else:
                    ax.plot3D(X[:, 0], X[:, 1], X[:,2], lstyle)
        
        if self.dim == 2:
            ax.set_aspect('equal')
            
        if self.dim == 3:
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
        
        """
        if show:
            plt.show()
        """
    def print_displacements(self):
        """ Prints nodal displacements
        """
        node_cnt = 0
        for node in self.nodes:
            print('*** Node {:d} ***'.format(node_cnt))
            node_cnt += 1
            print('u = {0:5.4f} mm'.format(node.u[0]))
            print('v = {0:5.4f} mm'.format(node.u[1]))
            print('phi = {0:5.4f} rad'.format(node.u[2]))


class Material:
    """ Class for material data
       
        For now, only linear elastic material is included.


        Parameters:
        -----------
        :param E: Young's modulus [MPa]
        :param nu: Poisson's ratio
        :param rho: density [kg / mm^3]
        :param G: shear modulus. If it is not given, it is calculated from
                    E and nu

        :type E: float
        :type nu: float
        :type rho: float
        :type G: float

        Variables:
        ----------
        :ivar young: Young's modulus [MPa]
        :ivar nu: Poisson's ratio
        :ivar density: density [kg / mm^3]
        :ivar shear_modulus: Shear modulus [MPa]

        :vartype young: float
        :vartype nu: float
        :vartype density: float
        :vartype shear_modulus: float
    """

    def __init__(self, E, rho, nu=None, G=None):
        """ Young's modulus """
        self.young = E
        """ Poisson ratio """
        self.nu = nu
        """ Density """
        self.density = rho
        
        """ Shear modulus """
        if G == None:
            G = 0.5*E/(1+nu)
        
        self.shear_modulus = G
        


class Section:
    """ Class for cross section properties

        All sections must have 'area', but specific profiles
        can have other properties as well
    
        Parameters:
        -----------
        :param A: cross-sectional area [mm^2]

        :type A: float

        Variables:
        ----------
        :ivar area: cross-sectional area [mm^2]

        :vartype area: float

    """

    def __init__(self, A):
        self.A = A
        """ cross-sectional area [mm^2] """


class BeamSection(Section):
    """ Class for beam element cross sections 

        Parameters:
        -----------
        :param A: cross-sectional area [mm^2]
        :param Iy: second moment of area with respect to major axis [mm^4]
        :param Iz: second moment of area with respect to minor axis [mm^4]
        :param J: torsional constant [mm^4]

        :type A: float
        :type Iy: float
        :type Iz: float
        :type J: float

        Variables:
        ----------
        :ivar Iy: second moment of area [mm^4]

        :vartype Iy: float
    """

    def __init__(self, A, Iy, Iz=None, J=None):
        Section.__init__(self, A)
        """ Second moment of area [mm^4]"""
        
        self.I = [0,0]
        self.I[0] = Iy
        
        if Iz is not None:
            self.I[1] = Iz
        
        #self.Iy = Iy
        #self.Iz = Iz
        """ Torsional constant """
        self.J = J
    
    @property
    def Iy(self):
        
        return self.I[0]

    @property
    def Iz(self):
        
        return self.I[1]

class Support:
    """ Supported nodes / degrees of freedom

        Parameters:
        -----------
        :param sid: support id
        :param node: supported node
        :param dof: degrees of freedom
        :param val: value for supports (default = 0.0, but theoretically
                    the displacement could have a non-zero value)

        :type sid: int
        :type node: FEMNode
        :type dof: list
        :type val: float

        Variables:
        -----------
        :ivar sid: support id
        :ivar node: supported node
        :ivar dof: degrees of freedom
        :ivar val: value for supports (default = 0.0, but theoretically
                    the displacement could have a non-zero value)

        :vqrtype sid: int
        :vqrtype node: FEMNode
        :vqrtype dof: list
        :vqrtype val: float

    """

    def __init__(self, sid, node, dof, val=0.0):
        self.sid = sid
        """ support id"""
        self.node = node
        """ supported node"""
        self.dof = dof
        """ degrees of freedom
            Alt. 1: dof = [1, 0,0 1, 0, 1] with ones for supported dofs
            Calling nodal_dofs() excludes the supported dofs
            
            Alt. 2: dof = [0, 3, 5] with indices of supported dofs
            Calling nodal_dofs() does not exclude the supported dofs but this
            is done in linear_statics()
        """
        
        #self.node.dofs = dof
        self.val = val
        """ value for supports"""


class Load:
    """ General class for loads

        Parameters:
        -----------
        :param sid: load id

        :type sid: int

        Variables:
        ----------
        :ivar sid: load id

        :vartype sid: int
    """

    def __init__(self, sid):
        self.sid = sid # Load id
        
        self.f = 1.0 # scaling factor
        

    @abstractclassmethod
    def load_and_dofs(self):
        """ Returns the loads to be inserted to the global load vector
            and the corresponding global degrees of freedom

            Returns:
            ---------
            :return: load to be inserted to the global load vector and
                    the corresponding global degrees of freedom
            :rtype: (int, int)
        """
        pass


class PointLoad(Load):
    """ Class for point loads (forces and moments)

        Parameters:
        ------------
        :param sid:  load id
        :param node: node subjected to the load
        :param v: load vector [Fx, Fy, Mz] [kN, kN, kNm]
        :param f: scaling factor

        :type sid: int
        :type node: FEMNode
        :type v: list
        :type f: float

        Variables:
        --------------
        :ivar sid:  load id
        :ivar node: node subjected to the load
        :ivar v: load vector [Fx, Fy, Mz] [kN, kN, kNm]
        :ivar f: scaling factor

        :vartype sid: int
        :vartype node: FEMNode
        :vartype v: list
        :vartype f: float

    """

    def __init__(self, sid, node, v, f):
        Load.__init__(self, sid)

        self.node = node
        """ node subjected to the load"""
        self.v = v
        """ load vector"""
        self.f = f
        """ scaling factor"""

    def load_and_dofs(self, order=1):
        """ Returns the loads to be inserted to the global load vector
            and the corresponding global degrees of freedom

            Returns:
            ---------
            :return: load to be inserted to the global load vector and
                    the corresponding global degrees of freedom
            :rtype: (int, int)

        """

        """ Load vector """
        F = self.f * np.array(self.v)

        
        """ Degrees of freedom """
        dofs = np.array(self.node.dofs)

        """ Find free degrees of freedom (those with numbering
                                          greater than -1) 
        """
        nzero_dofs = dofs >= 0

        """ Get free degrees of freedom and the corresponding part
            of the load vector
        """
        d = dofs[nzero_dofs]
        Fnz = F[nzero_dofs]

        return Fnz, d


class LineLoad(Load):
    """ Class for 2D line loads (forces and moments)

        Parameters:
        -----------
        :param sid: Load id
        :param eid: Element subjected to load
        :param xloc: Starting and ending locations of the load (local coordinates) [loc1, loc2]
        :param qval: Value of the load at the coordinates xloc [kN/m] [q1, q2]
        :param direction: Direction of load in global coordinates, 0: x, 1: y
        :param coords: (Optional, default=1) Coordinates used, 0: local, 1: global

        :type sid: int
        :type eid: Element
        :type xloc: list
        :type qval: list
        :type direction: int
        :type coords: int

    """

    def __init__(self, sid, eid, xloc, qval, direction, coords='global'):

        Load.__init__(self, sid)

        self.elem = eid
        """ Elemnt subjected to load"""
        self.xloc = xloc
        """ Starting and ending locations of the load in local coordinates"""
        self.qval = qval
        """ Values of the load at the coordinates xloc"""
        self.dir = direction
        """ Direction of the load, 0: x, 1: y"""
        self.coords = coords
        """ Coordinates used, 0: local, 1: global (Default=1)"""

    def load_and_dofs(self, order=1):
        """ Returns the loads to be inserted to the global load vector
            and the corresponding global degrees of freedom

            Returns:
            ---------

            :return: load to be inserted to the global load vector and
                    the corresponding global degrees of freedom
            :rtype: (int, int)

        """

        # Get load vector
        if self.dir == "x":
            q = np.array([self.f*self.qval[0], 0, 0])
        else:
            q = np.array([0, self.f*self.qval[0], 0])

        # Get equivalent nodal loads   
        if isinstance(self.elem,list):
            print(self.elem)
        F = self.elem.equivalent_nodal_loads(self, order)
        
        # Number of degrees of freedom
        dofs = np.array(self.elem.global_dofs())

        # Find free degrees of freedom (those with numbering greater than -1)
        nzero_dofs = dofs >= 0

        # Get free degrees of freedom and the corresponding part
        #    of the load vector

        d = dofs[nzero_dofs]
        Fnz = F[nzero_dofs]

        return Fnz, d


class LoadCase:
    """ Class for storing load case data

        Parameters:
        -----------
        :param support: Support id, supports having this id are added to the calculation model
        :param load: Load id, loads having this id are added to the calculation model

        :type support: int
        :type load: int

        Variables:
        -----------
        :ivar support: Supports id, supports having this id are added to the calculation model
        :ivar load: Loads id, loads having this id are added to the calculation model

        :vartype support: int
        :vartpe load: int

    """

    def __init__(self, support, load):
        self.support = support
        self.load = load


class FEMNode:
    """ Class for FEM nodes.

        Parameters:
        -----------
        :param nid: node id (integer)
        :param x: x-coordinate of the node
        :param y: y-coordinate of the node
        :param z: z-coordinate of the node (optional)

        :type x: float
        :type y: float
        :type z: float

        Variables:
        ----------
        :ivar x: Coordinates of the node
        :ivar dofs: Degrees of freedom for node [0, 1, 2] [Ux, Uy, Rz]
                In 3D: [Ux, Uy, Uz, Rx, Ry, Rz]
        :ivar u: Nodal displacements due to linear static analysis
        :ivar v: Nodal displacements due to linear buckling analysis

        :vartype x: np.array
        :vartype dofs: list
        :vartype u: np.array
        :vartype v: np.array


    """

    def __init__(self, nid, x, y, z=None):
        """ Constructor
            If z-coordinate is given, then the problem becomes immediately 3D
        """

        """ Node id """
        self.nid = nid
        

        """ Node's degrees of freedom """
        # NOTE: for multiple load cases, the dimension of u must be increased for each load case
        #self.u = np.array([])
        self.u = {}
        self.u2 = {}
        
        """ Nodal coordinates"""
        if z is not None:
            self.coord = np.array([x, y, z])
            self.X = np.array([x, y, z])
            # Degrees of freedom (integer values)
            #self.dofs = [0, 0, 0, 0, 0, 0]
            self.dofs = -np.ones(6)
            # Nodal displacements
            # [Ux, Uy, Uz, Rx, Ry, Rz]
            self.v = np.array([0, 0, 0, 0, 0, 0])
        else:
            self.coord = np.array([x, y])
            self.X = np.array([x, y])
            self.dofs = np.array([-1, -1, -1])
            # [Ux, Uy, Uz]
            self.v = np.array([0,0,0])
        
        
        """ Nodal displacement vector (linear buckling)"""
        # List of FrameMember-type objects that are connected to this node
        self.parents = []

    @property
    def x(self):
        return self.coord[0]

    @x.setter
    def x(self, val):
        self.coord[0] = val

    @property
    def y(self):
        return self.coord[1]

    @y.setter
    def y(self, val):
        self.coord[1] = val

    @property
    def z(self):
        if len(self.coord) == 2:
            return None
        return self.coord[2]

    @z.setter
    def z(self, val):
        self.coord[2] = val

    def __repr__(self):
        return f'FEMNode(nid={self.nid}, x={self.x}, y={self.y}, z={self.z})'

    def plot(self, print_text='', c='k', axes=None, style='.r'):

        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes

        if self.z is None:
            #ax.plot(X[:, 0], X[:, 1], lstyle, color=(0.6,0.6,0.6))
            ax.plot(self.x, self.y, style)
            ax.text(self.x, self.y, print_text)
        else:
            ax.scatter3D(self.x, self.y, self.z, style)
            ax.text(self.x, self.y, self.z, print_text)

class Element(metaclass=ABCMeta):
    """ Class for 1D finite elements: bars and beams

        Parameters:
        -----------
        :param n1: Start node
        :param n2: End node
        :param section: Element's section
        :param material: Element's material

        :type n1: FEMNode
        :type n2: FEMNode
        :type section: Section
        :type material: Material

        Variables:
        -----------
        :ivar nodes: Element's nodes
        :ivar material: Element's material
        :ivar section: Element's section
        :ivar axial_force: Axial force on element's nodes
        :ivar shear_force: Shear force on element's nodes
        :ivar bending_moment: Bending moment on element's nodes
        :ivar floc: Local loads

        :vartype nodes: list
        :vartype material: Material
        :vartype section: Section
        :vartype fint: dict .. internal forces
        :vartype axial_force: list
        :vartype shear_force: list
        :vartype bending_moment: list
        :vartype floc: np.array

    """

    def __init__(self, n1, n2, section, material):
        self.nodes = [n1, n2]
        self.material = material        
        self.section = section
        self.fint = {}
        self.fint2 = {}
        #self.axial_force = [0.0, 0.0]
        #self.shear_force = [0.0, 0.0]
        #self.bending_moment = [0.0, 0.0]
        self.floc = {} #np.array([])
        
        """ Releases is a list of degrees of freedom for those
            internal forces that should be set to zero
        """
        self.releases = []
        
        """ Dimension of the element """
        self.dim = 2
        
        #self.x = self.coord()
        #self.len = self.length()
        #self.dcos = self.direction_cosines()

    def coord(self,initial=True):
        """ Nodal coordinates of the element
            
            input: initial .. True, if initial coordinates are given
                                False, is current coordinates are given.

            Returns:
            --------
            :return: Nodal coordinates of the element
            :rtype: np.array

        """
        if initial:
            X = np.array([self.nodes[0].coord, self.nodes[1].coord])
        else:
            X = np.array([self.nodes[0].X, self.nodes[1].X])
        return X
    
    def direction_cosines(self,initial=True):
        """ Calculates element's direction cosines

            Returns:
            --------
            :return: Element's direction cosines
            :rtype: np.array

        """
        
        
        X = self.coord(initial)
        dX = X[1, :] - X[0, :]
        L = self.length(initial)
        c = dX / L
        return c
        
        #return (self.x[1]-self.x[0])/self.len

    @abstractclassmethod
    def equivalent_nodal_loads(self, q):
        """ Compute equivalent nodal loads with line loads in
            vector q
        """
        pass

    def length(self,initial=True):
        """ Member length

            Returns:
            --------
            :return: length of the element
            :rtype: float

        """
        
        #X = self.__x
        X = self.coord(initial)
        L = np.linalg.norm(X[1] - X[0])        
        return round(L, 3)
        #return round(np.linalg.norm(self.x[1]-self.x[0]),3)

    @abstractclassmethod
    def local_stiffness_matrix(self, E, A, I, L):
        """ Generates stiffness matrix in local coordinates

            Returns:
            --------
            :return: element's stiffness matrix in local coordinates
            :rtype: np.array

        """
        pass

    @abstractclassmethod
    def local_geometric_stiffness_matrix(self):
        """ Geometric stiffness matrix in local coordinates

            Returns:
            ---------
            :return: element's geometric stiffness matrix in local coordinates
            :rtype: np.array
        """

    @abstractclassmethod
    def stiffness_matrix(self):
        """ Computes element's global stiffness matrix

            Returns:
            ---------
            :return: element's stiffness matrix in global coordinates
            :rtype: np.array

        """
        pass

    def init_dofs(self):
        """ Initialize dofs 
            Does nothing for beam elements
        """
        pass

    @abstractclassmethod
    def global_dofs(self):
        """ Get numbering of element's global degrees of freedom

            Returns:
            --------
            :return: list of end nodes dofs added together
            :rtype: list
        """
        pass

    def nodal_displacements(self,lcase=0,order=1):
        """ Get nodal displacements of an element in global coordinates
            Requires previously performed structural analysis such
            that nodal displacements are available.

            Returns:
            --------
            :return: Element's nodal displacements
            :rtype: np.array
        """
        
        if order == 1:
            q = np.concatenate((self.nodes[0].u[lcase],
                                   self.nodes[1].u[lcase]))
        else:
            q = np.concatenate((self.nodes[0].u2[lcase],
                                   self.nodes[1].u2[lcase]))

        
        return q
    
    def local_displacements(self, lcase=0, order=1, initial=True):
        """ Nodal displacements in local coordinates

            Returns:
            ---------
            :return: Nodal displacements in local coordinates
            :rtype: np.array

        """
        T = self.transformation_matrix(initial)
        q = self.nodal_displacements(lcase,order)

        return T.dot(q)

    def add_loadcase(self,lcase=0):
        """ Creates a place holder for internal forces 
            fx .. axial force
            fy .. shear force with respect to horizontal axis
            fz .. shear force with respect to vertical axis
            mx .. torsion
            my .. bending moment with respect to horizontal axis
            mz .. bending moment with respect to vertical axis
            
            In 2d, the following internal forces are present
            fx, fz, my
        """
        
        self.fint[lcase] = {'fx': [0,0], 'fy': [0,0], 'fz': [0,0],\
                            'mx': [0,0], 'my': [0,0], 'mz': [0,0]}
        
        self.fint2[lcase] = {'fx': [0,0], 'fy': [0,0], 'fz': [0,0],\
                            'mx': [0,0], 'my': [0,0], 'mz': [0,0]}
            
    @abstractclassmethod
    def transformation_matrix(self):
        """ Calculates transformation matrix from local to global coordinates

            Returns:
            --------
            :return: element's transformation matrix
            :rtype: np.array

         """
        pass

    @abstractclassmethod
    def internal_forces(self):
        """ Calculates internal forces and saves them to element's attributes

            axial_force = [n1, n2]-- axial force on node1 and node2
            shear_force = [n1, n2] -- shear force on node1 and node2
            bending_moment = [n1, n2] -- bending moment on node1 and node2
        """
        pass
    
    def plot(self, print_text='', c='k', axes=None, lstyle='-',
             deformed=None,buckling_mode=None,scale=1.0):

        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes

        X = self.coord()
        
        # Plot members
        #ax.plot([X[0][0], X[1][0]], [X[0][1], X[1][1]], c)
        # Plot text

        Xmid = X[0, :] + 0.5 * (X[1, :] - X[0, :])            
            
        if self.dim == 2:
            #ax.plot(X[:, 0], X[:, 1], lstyle, color=(0.6,0.6,0.6))
            if deformed is not None:                
                lstyle = '--'
                dstyle = '-'
                c = (0.6,0.6,0.6)
                Xd = deepcopy(X)
                u = np.array([n.u[deformed][:2] for n in self.nodes])
                
                Xd += scale*u
                
                ax.plot(Xd[:, 0], Xd[:, 1], dstyle, color='k')
                
            # Plot initial element
            ax.plot(X[:, 0], X[:, 1], lstyle, color=c)
                
            ax.text(Xmid[0], Xmid[1], print_text)
        else:
            ax.plot3D(X[:, 0], X[:, 1], X[:,2], lstyle)
            ax.text(Xmid[0], Xmid[1], Xmid[2], print_text)

        """
        if self.mtype == 'beam':
            horzalign = 'center'
            vertalign = 'bottom'

        elif self.mtype == 'column':
            horzalign = 'right'
            vertalign = 'center'

        else:
            horzalign = 'center'
            vertalign = 'center'

        x, y = self.to_global(0.3) - self.perpendicular * 50
        rot = np.degrees(self.angle)

        if print_text:
            ax.text(x, y, str(self.mem_id) + ": " + str(self.cross_section),
                     rotation=rot, horizontalalignment=horzalign,
                     verticalalignment=vertalign)
        else:
            ax.text(x, y, str(self.mem_id),
                     rotation=rot, horizontalalignment=horzalign,
                     verticalalignment=vertalign)
        """


#from metku.frame2d import *
from metku.raami.raami import *
from metku.sections.steel.RHS import RHS

f = Raami()

L = 5000.0
H = 2000.0
nodes = [[0.0, 0.0], [0.0, H], [L, H], [L, 0.0]]

s = RHS(200, 200, 6)

for node in nodes:
    f.add(FrameNode(node))

f.add(SteelFrameMember(nodes=f.nodes[:2], section=s, mem_type='column', nel=6))
f.add(SteelFrameMember(nodes=f.nodes[1:3], section=s, mem_type='beam', nel=6, hinges=[True, True]))
f.add(SteelFrameMember(nodes=f.nodes[2:4], section=s, mem_type='column', nel=6))

f.add(FixedSupport(f.nodes[0]))
f.add(FixedSupport(f.nodes[-1]))

f.add(LineLoad(f.members[1], [-20, -20], 'y', coord_sys='local', load_id=0, ltype='snow', name='Lumi'))
f.add(LineLoad(f.members[0], [3, 3], 'x', coord_sys='global', load_id=1, ltype='wind', name='Tuuli x+'))

f.add(LoadCombination(comb_id=2, comb_type='ULS', load_cases=list(f.load_cases.values())))

f.load_combs[2].combine_loads()
f.plot()
f.generate_fem()
f.structural_analysis(load_id=2, support_method="REM")
f.bmd(scale=10, load_id=2)

#zelf toegoevoegd:
f.design_members(load_id=2)
f.print_member_utilization()

N = f.members[0].nodal_forces[2]['N']
Vz = f.members[0].nodal_forces[2]['Vz']
My = f.members[0].nodal_forces[2]['My']
print(N, Vz, My)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
try:
    from metku.frame2d.frame2d import SteelBeam
    from metku.sections.steel import SteelSection
    from metku.optimization.structopt import *
    from metku.sections.steel.catalogue import ipe_profiles, shs_profiles, hea_profiles
    from metku.sections.steel import ISection, WISection, RHS
except:
    from frame2d.frame2d import SteelBeam
    from sections.steel import SteelSection
    from optimization.structopt import *
    from sections.steel.catalogue import ipe_profiles, shs_profiles, hea_profiles
    from sections.steel import ISection, WISection, RHS


class StructuralProblem(OptimizationProblem):
    """
    Class for general structural problem

    """

    def __init__(self,
                 name="Structural Optimization Problem",
                 structure=None,
                 var_groups=None,
                 con_groups=None,
                 constraints=None,
                 profiles=None,
                 obj=None):
        """ Constructor 
            Parameters:
            ------------
            :param name: Name of the problem (tuple)
            :param structure: Structure to be optimized (Frame2D member?)
            :param var_groups: grouping of design variables
            :param con_groups: constraint groups
            :param constraints: constraints of the problem
            :param profiles: available profiles
            :param obj: objective function
    
            :type name: string
            :type structure: Frame2D member?
            :type var_groups: list or dict?
            :type con_groups: list or dict?
            :type constraints: list or dict?
            :type profiles: list or dict?
            :type obj: callable or string?
    
    
            Variables:
            ----------
            :ivar vars: list of variables
            :ivar var_groups:
            :ivar con_groups:
            :ivar profiles:
            :ivar constraints:
            :ivar
        
        """
        super().__init__(name=name,
                         structure=structure)
        self.vars = []
        self.var_groups = var_groups
        self.con_groups = con_groups
        if profiles is None:
            profiles = list(ipe_profiles.keys())
        self.profiles = profiles
        
        """ Generate variables.
            By default, index variables are generated. This can be overridden
            by var_groups, which gives the variable data.
        """
        self.create_variables()
        
        """ Generate constraints 
            By default, cross-section resistance constraints are generated.
            This can be overridden by 'constraints' dict.
        """
        if constraints is None:
            constraints = {
                'compression': True,
                'tension': True,
                'shear': True,
                'bending_moment': True
            }
        self.constraints = constraints
        self.create_constraints(**self.constraints)
        self.group_constraints()
        if obj is not None:
            self.obj = obj
        else:
            self.create_objective()

    @classmethod
    def available_constraints(self):
        """
        Returns all available constraints and their corresponding keyword
        :return:
        """
        return dict(joint_geometry_constraints=False,
                    joint_strength_constraints=False,
                    compression=False,
                    tension=False,
                    shear=False,
                    bending_moment=False,
                    buckling_y=False,
                    buckling_z=False,
                    LT_buckling=False,
                    compression_bending_y=False,
                    compression_bending_z=False,
                    deflection_y=None,
                    deflection_x=None,
                    web_class=None,
                    flange_class=None,
                    alpha_cr=None)

    def joint_geometry_constraints(self, joint):
        """
        Creates joint geometry constraint functions
        :param joint: TrussJoint
        :return: constraint functions in a dict
        """

        con_funs = {}

        if joint.joint_type == 'Y':
            w1 = list(joint.webs.values())[0]

            def b(x):
                """
                Web's breadth constraint
                """
                return w1.cross_section.B / joint.chord.cross_section.B - 1

            def b_min(x):
                """
                Minimum breadth for web 1
                """
                val = 0.35 * joint.chord.cross_section.B
                val /= w1.cross_section.B
                val -= 1
                return val

            def theta_min(x):

                val = 30 / np.degrees(joint.theta1) - 1
                # val = 30 - np.degrees(w1.angle - joint.chord.angle)
                return val

            con_funs = {
                f'Joint {joint.jid} b': b,
                f'Joint {joint.jid} b1_min': b_min,
                f'Joint {joint.jid} theta1_min': theta_min,
            }

        elif joint.joint_type == 'K':
            w1, w2 = joint.webs.values()

            def beta(x):
                """
                Beta constraint
                """
                b1 = w1.cross_section.B
                b2 = w2.cross_section.B
                h1 = w1.cross_section.H
                h2 = w2.cross_section.H
                b0 = joint.chord.cross_section.B

                return (b1 + b2 + h1 + h2) / (4 * b0) - 1
                # return (b1 + b2 + h1 + h2) - (4 * b0)

            def b1b0(x):
                b1 = w1.cross_section.B
                b0 = joint.chord.cross_section.B
                t0 = joint.chord.cross_section.T

                return (0.1 + 0.01 * b0 / t0) / (b1 / b0) - 1

            def b2b0(x):
                b2 = w2.cross_section.B
                b0 = joint.chord.cross_section.B
                t0 = joint.chord.cross_section.T

                return (0.1 + 0.01 * b0 / t0) / (b2 / b0) - 1

            def g_min(x):
                """
                Minimum value for gap
                """
                gap = w1.cross_section.T + w2.cross_section.T
                val = gap - joint.g1
                val /= abs(joint.g1)
                return val

            def g_min_beta(x):
                """
                Minimum value for gap
                """
                beta = joint.rhs_joint.beta()
                val = 0.5 * (1 - beta) * joint.chord.cross_section.B
                val /= abs(joint.g1)
                val -= 1
                return val

            def g_max_beta(x):
                """
                Maximum value for gap
                """
                beta = joint.rhs_joint.beta()
                val = joint.g1
                val /= (1.5 * (1 - beta) * joint.chord.cross_section.B)
                val -= 1
                return val

            def b1_min(x):
                """
                Minimum breadth for web 1
                """
                val = 0.35 * joint.chord.cross_section.B
                val /= w1.cross_section.B
                val -= 1
                return val

            def b1_max(x):
                val = -0.85 * joint.chord.cross_section.B
                val /= w1.cross_section.B
                val += 1
                return val

            def b2_min(x):
                val = 0.35 * joint.chord.cross_section.B
                val /= w2.cross_section.B
                val -= 1
                return val

            def b2_max(x):
                val = -0.85 * joint.chord.cross_section.B
                val /= w2.cross_section.B
                val += 1
                return val

            def theta1_min(x):

                val = 30 / np.degrees(joint.theta1) - 1
                # val = 30 - np.degrees(w1.angle - joint.chord.angle)
                return val

            def theta1_max(x):
                val = np.degrees(joint.theta1) / 90 - 1
                # val = 30 - np.degrees(w1.angle - joint.chord.angle)

                return val

            def theta2_min(x):
                val = 30 / np.degrees(joint.theta2) - 1
                # val = 30 - np.degrees(w2.angle - joint.chord.angle)
                return val

            def theta2_max(x):
                val = np.degrees(joint.theta2) / 90 - 1
                return val

            def e_pos(x):
                return joint.e / ((0.25 * joint.chord.cross_section.H)) - 1


            def e_neg(x):
                return -1 - joint.e / (0.55 * joint.chord.cross_section.H)

            con_funs = {
                f'Joint {joint.jid} beta': beta,
                f'Joint {joint.jid} b1b0': b1b0,
                f'Joint {joint.jid} b2b0': b2b0,
                f'Joint {joint.jid} g_min': g_min,
                f'Joint {joint.jid} g_min_beta': g_min_beta,
                f'Joint {joint.jid} g_max_beta': g_max_beta,
                f'Joint {joint.jid} b1_min': b1_min,
                f'Joint {joint.jid} b1_max': b1_max,
                f'Joint {joint.jid} b2_min': b2_min,
                f'Joint {joint.jid} b2_max': b2_max,
                f'Joint {joint.jid} theta1_min': theta1_min,
                f'Joint {joint.jid} theta1_max': theta1_max,
                f'Joint {joint.jid} theta2_min': theta2_min,
                f'Joint {joint.jid} theta2_max': theta2_max,
                f'Joint {joint.jid} e_pos': e_pos,
                f'Joint {joint.jid} e_neg': e_neg}

        elif joint.joint_type == 'KT':
            # TODO
            pass

        return con_funs

    def joint_strength_constraints(self, joint):
        """
        Creates joint strength costraint functions

        :param joint: TrussJoint to be calculated
        :return: constraint functions in a dict
        """
        con_funs = {}

        if joint.joint_type == 'Y':
            w1 = list(joint.webs.values())[0]

            def chord_face_failure(x):
                NRd1 = joint.rhs_joint.chord_face_failure()
                return w1.ned / NRd1 - 1

            def chord_web_buckling(x):
                NRd1 = joint.rhs_joint.chord_web_buckling()
                return -w1.ned / NRd1 - 1

            def brace_failure(x):
                N1Rd = joint.rhs_joint.brace_failure()
                return abs(w1.ned) / N1Rd - 1

            con_funs = {
                f'Joint {joint.jid} chord face failure': chord_face_failure,
                f'Joint {joint.jid} chord web buckling': chord_web_buckling,
                f'Joint {joint.jid} brace failure': brace_failure,
            }


        elif joint.joint_type == 'K':
            w1, w2 = joint.webs.values()

            def chord_face_failure_1(x):

                NRd1, NRd2 = joint.rhs_joint.chord_face_failure()
                return abs(w1.ned) / NRd1 - 1
                # return abs(w1.ned) - NRd1

            def chord_face_failure_2(x):
                NRd1, NRd2 = joint.rhs_joint.chord_face_failure()
                return abs(w2.ned) / NRd2 - 1
                # return abs(w2.ned) - NRd2

            def punching_shear_1(x):
                NRd1 = joint.rhs_joint.punching_shear()[0]
                return abs(w1.ned) / NRd1 - 1
                # return abs(w1.ned) - NRd1

            def punching_shear_2(x):
                NRd2 = joint.rhs_joint.punching_shear()[1]
                return abs(w2.ned) / NRd2 - 1
                # return abs(w2.ned) - NRd2

            def chord_shear_0(x):
                joint.rhs_joint.V0 = joint.V0
                (N1Rd, N2Rd), N0Rd = joint.rhs_joint.chord_shear()

                return joint.N0 / N0Rd - 1
                # return joint.N0 - N0Rd

            def chord_shear_1(x):
                joint.rhs_joint.V0 = joint.V0
                (N1Rd, N2Rd), N0Rd = joint.rhs_joint.chord_shear()

                return abs(w1.ned) / N1Rd - 1

            def chord_shear_2(x):
                joint.rhs_joint.V0 = joint.V0
                (N1Rd, N2Rd), N0Rd = joint.rhs_joint.chord_shear()

                return abs(w2.ned) / N2Rd - 1

            def brace_failure_1(x):
                N1Rd = joint.rhs_joint.brace_failure()[0]
                return abs(w1.ned) / N1Rd - 1
                # return abs(w1.ned) - N1Rd

            def brace_failure_2(x):
                N2Rd = joint.rhs_joint.brace_failure()[1]
                return abs(w2.ned) / N2Rd - 1
                # return abs(w2.ned) - N2Rd

            con_funs = {
                f'Joint {joint.jid} chord face failure 1': chord_face_failure_1,
                f'Joint {joint.jid} chord face failure 2': chord_face_failure_2,
                f'Joint {joint.jid} punching shear 1': punching_shear_1,
                f'Joint {joint.jid} punching shear 2': punching_shear_2,
                f'Joint {joint.jid} chord shear 0': chord_shear_0,
                f'Joint {joint.jid} chord shear 1': chord_shear_1,
                f'Joint {joint.jid} chord shear 2': chord_shear_2,
                f'Joint {joint.jid} brace failure 1': brace_failure_1,
                f'Joint {joint.jid} brace failure 2': brace_failure_2,
            }

        elif joint.joint_type == 'KT':
            # TODO
            pass

        return con_funs

    def cross_section_constraints(self, mem, elem):
        """
        Creates cross-section constraints

        :param mem: FrameMember object
        :param elem: Element object

        :return: dict of created functions
        """

        def compression(x):
            N = elem.axial_force[0]
            return -N / mem.NRd - 1

        def tension(x):
            N = elem.axial_force[0]
            return N / mem.NRd - 1

        def shear(x):
            V = elem.shear_force[0]
            return abs(V) / mem.VRd - 1

        def bending_moment(x):
            # Moment about y
            # TODO: Moment about z
            M = elem.bending_moment[0]
            return abs(M) / mem.MRd[0] - 1

        cons = {
            'compression': compression,
            'tension': tension,
            'shear': shear,
            'bending_moment': bending_moment
        }

        return cons

    def stability_constraints(self, mem, section_class):
        """
        Creates stability constraint functions

        :param mem: FrameMember object
        :return:
        """

        def buckling_y(x):
            return -min(mem.ned) / mem.NbRd[0] - 1

        def buckling_z(x):
            return -min(mem.ned) / mem.NbRd[1] - 1

        def com_compression_bending_y(x):
            return mem.check_beamcolumn(section_class=section_class)[0] - 1

        def com_compression_bending_z(x):
            return mem.check_beamcolumn(section_class=section_class)[1] - 1

        def lt_buckling(x):
            return abs(max(mem.myed)) / mem.MbRd - 1

        cons = {
            'buckling_y': buckling_y,
            'buckling_z': buckling_z,
            'compression_bending_y': com_compression_bending_y,
            'compression_bending_z': com_compression_bending_z
        }

        if isinstance(mem.profile, RHS):
            cons['lt_buckling'] = lt_buckling

        return cons

    def deflection_constraints(self, mem, limit_x=None, limit_y=None):
        """
        Creates deflection constraint functions
        :param mem: FrameMember object
        :return:
        """

        cons = {}

        def deflection_y(x):
            displacements = mem.nodal_displacements.values()
            vals = [abs(l[0][1]) for l in displacements]
            abs_max = max(vals)
            return abs_max / limit_y - 1

        def deflection_x(x):
            displacements = mem.nodal_displacements.values()
            vals = [abs(l[0][0]) for l in displacements]
            abs_max = max(vals)
            return abs_max / limit_x - 1

        if limit_x:
            cons['deflection_x'] = deflection_x
        if limit_y:
            cons['deflection_y'] = deflection_y

        return cons

    def cross_section_class_constraints(self, mem, column_class=3,
                                        tube_class=2):
        """
        Cross-section class constraints
        :param mem: FrameMember -object
        :return:
        """
        K_web = [0,33,38,42, np.inf]
        K_flange = [0,9,10,14, np.inf]

        def web_class(x):
            if isinstance(mem.cross_section, (ISection, WISection)):
                c = mem.cross_section.cw
                t = mem.cross_section.tw
                K = K_web[column_class]

            elif isinstance(mem.cross_section, RHS):
                c = mem.cross_section.c_web
                t = mem.cross_section.T
                K = K_web[tube_class]

            return c / t - K * mem.cross_section.eps

        def flange_class(x):
            if isinstance(mem.cross_section, (ISection, WISection)):
                c = mem.cross_section.cf_top
                t = mem.cross_section.tt
                K = K_flange[column_class]

            elif isinstance(mem.cross_section, RHS):
                c = mem.cross_section.c_flange
                t = mem.cross_section.T
                K = K_web[tube_class]

            return c / t - K * mem.cross_section.eps

        cons = {'web_class': web_class,
                'flange_class': flange_class
                }

        return cons

    def create_constraints(self,
                           joint_geometry_constraints=False,
                           joint_strength_constraints=False,
                           compression=False,
                           tension=False,
                           shear=False,
                           bending_moment=False,
                           buckling_y=False,
                           buckling_z=False,
                           lt_buckling=False,
                           compression_bending_y=False,
                           compression_bending_z=False,
                           deflection_y=None,
                           deflection_x=None,
                           web_class=None,
                           flange_class=None,
                           alpha_cr=None,
                           members=None
                           ):
        """
        Creates constraints

        NOTE! The name of these kwargs must match the key in the cons -dict
        created in constraint function creator method
        e.g.
        CORRECT:
        cons = {'compression': comp_fun, 'tension': ten_fun}
        FALSE:
        cons = {'comp_func': comp_fun, 'tension_con': ten_fun}

        :param compression:
        :param tension:
        :param shear:
        :param bending_moment:
        :param buckling_y:
        :param buckling_z:
        :param LT_buckling:
        :param compression_bending_y:
        :param compression_bending_z:
        :param deflection_y:
        :param deflection_x:
        :param alpha_cr:
        :return:
        """

        if members is None:
            members = self.structure.members.values()
            # Clear constraint list
            self.cons.clear()

        for mem in members:
            # STABILITY CONSTRAINTS
            for i, smem in enumerate(mem.members):
                if isinstance(mem.cross_section, WISection):
                    section_class = 3
                else:
                    section_class = 2
                stability_cons = self.stability_constraints(smem, section_class)
                for key, val in stability_cons.items():
                    if eval(key.lower()):
                        con = NonLinearConstraint(
                            name=f"{key}: {mem.mem_id}|{i}",
                            con_fun=val,
                            con_type='<',
                            fea_required=True,
                            vars=mem)
                        self.add(con)

            # CROSS-SECTION CONSTRAINTS
            for eid, elem in mem.elements.items():
                section_cons = self.cross_section_constraints(mem, elem)
                for key, val in section_cons.items():
                    if eval(key.lower()):
                        con = NonLinearConstraint(
                            name=f"{key}: {mem.mem_id}|{eid}",
                            con_fun=val,
                            con_type='<',
                            fea_required=True,
                            vars=mem)
                        self.add(con)

            # CROSS-SECTION CLASS CONSTRAINTS
            class_cons = self.cross_section_class_constraints(mem)
            for key, val in class_cons.items():
                if eval(key.lower()):
                    con = NonLinearConstraint(
                        name=f"{key}: {mem.mem_id}",
                        con_fun=val,
                        con_type='<',
                        fea_required=False,
                        vars=mem)
                    self.add(con)

            # DEFLECTION CONSTRAINTS
            def_cons = self.deflection_constraints(mem, deflection_x,
                                                   deflection_y)
            for key, val in def_cons.items():
                if eval(key.lower()):
                    con = NonLinearConstraint(
                        name=f"{key}: {mem.mem_id}",
                        con_fun=val,
                        con_type='<',
                        fea_required=True,
                        vars=mem)
                    self.add(con)

        # JOINTS
        for truss in self.structure.truss:
            for joint in truss.joints.values():
                # JOINT GEOMETRY
                if joint_geometry_constraints:
                    joint_geom_cons = self.joint_geometry_constraints(joint)
                    for key, val in joint_geom_cons.items():
                        con = NonLinearConstraint(
                            name=f"{key}:",
                            con_fun=val,
                            con_type='<',
                            fea_required=False)
                        self.add(con)
                # JOINT STRENGTH
                if joint_strength_constraints:
                    joint_str_cons = self.joint_strength_constraints(joint)
                    for key, val in joint_str_cons.items():
                        con = NonLinearConstraint(
                            name=f"{key}:",
                            con_fun=val,
                            con_type='<',
                            fea_required=True)
                        self.add(con)


    def create_variables(self):
        """
        Creates design variables
        """
        self.vars = []

        # If no grouping is given, create IndexVariables
        # with all IPE and HEA profiles
        # TODO
        if self.var_groups is None:
            self.var_groups = []
            for mem in self.structure.members.values():
                group = {
                    "name": f"{mem.mtype.capitalize()} {mem.mem_id}",
                    "objects": [mem],
                    "value": 0,
                    "profiles": self.profiles,
                    "property": "profile",
                    "var_type": 'index',
                }
                self.var_groups.append(group)

        # Main loop
        for i, group in enumerate(self.var_groups):
            if group["var_type"] == 'discrete':
                var = DiscreteVariable(
                    name=group['name'],
                    values=group['values'],
                    value=group['value'],
                    target={"property": group["property"],
                            "objects": group["objects"]}
                )
                self.add(var)

            elif group["var_type"] == 'continuous':
                # Multiple properties
                if "properties" in group.keys():
                    bounds = group["bounds"]
                    if len(bounds) != len(group["properties"]):
                        raise ValueError(
                            "There must be same number of bounds as properties!"
                            f"{group['bounds']} != {group['properties']}")
                    for bounds, prop, value in zip(group["bounds"],
                                                   group["properties"],
                                                   group["values"]):
                        lb, ub = bounds
                        var = Variable(
                            name=group['name'],
                            lb=lb,
                            ub=ub,
                            value=value,
                            target={"property": prop,
                                    "objects": group["objects"]}
                        )
                        self.add(var)

                # Single property
                else:
                    var = Variable(
                        name=group['name'],
                        lb=group['lb'],
                        ub=group['ub'],
                        value=group['value'],
                        target={"property": group["property"],
                                "objects": group["objects"]}
                    )
                    self.add(var)

            elif group["var_type"] == 'index':
                var = IndexVariable(
                    name=group['name'],
                    value=group['value'],
                    values=group['values'],
                    target={"property": group["property"],
                            "objects": group["objects"]}
                )
                self.add(var)

            elif group["var_type"] == 'binary':
                # TODO
                pass
            else:
                raise ValueError("var_type must be either 'discrete',"
                                 " 'continuous', 'index' or 'binary")

    def index_to_binary(self, params=('H', 'B', 'T'),
                        chord_profiles=shs_profiles.keys(),
                        web_profiles=shs_profiles.keys(),
                        col_profiles=hea_profiles.keys()):
        """
        Creates binary variables and continuous variables
        :return:
        """
        idx_vars = [var for var in self.vars if isinstance(var, IndexVariable)]
        # Temp member instance to get needed values
        mem = SteelBeam([[0, 0], [1000, 0]])

        for var in idx_vars:
            if isinstance(var.target['objects'][0], SteelSection):
                new_var = DiscreteVariable(
                    name=f"{var.name}",
                    values=var.values,
                    value=var.values[var.value],
                    id=var.id,
                    target=var.target
                )
                # Add new variable
                self.add(new_var)
            else:
                for param in params:
                    values = []
                    if var.target['objects'][0].mtype == 'web':
                        profiles = web_profiles
                    elif var.target['objects'][0].mtype == 'column':
                        profiles = col_profiles
                    elif 'chord' in var.target['objects'][0].mtype:
                        profiles = chord_profiles
                    else:
                        break
                    for profile in profiles:
                        mem.profile = profile
                        values.append(mem.cross_section.__getattribute__(param))
                    new_var = DiscreteVariable(
                        name=f"{var.name} {param}",
                        values=values,
                        value=values[min(var.value, len(values) - 1)],
                        id=var.id,
                        target={"property": param,
                                "objects": [obj.cross_section for obj in
                                            var.target["objects"]]}
                    )
                    # Add new variable
                    self.add(new_var)
                # Remove index variable
            self.all_vars.remove(var)

    def group_constraints(self):

        if self.con_groups is not None:
            con_groups = [group for group in self.con_groups if
                          "constraints" in group]

            for group in self.con_groups:
                cons = group['constraints']
                for obj in group['objects']:
                    for con in self.cons:
                        if obj in con.vars:
                            # Remove old constraints
                            self.cons.remove(con)
                            del con
                    init_cons = self.constraints.copy()
                    init_cons.update(cons)
                    self.create_constraints(**init_cons,
                                            members=group['objects'])

    def create_objective(self):
        """ By default, weight of the structure is taken as the
            objective function.
        """
        def obj_fun(x):
            """ Variable substitution is not shown """
            return self.structure.weight

        obj = ObjectiveFunction(name="Weight",
                                obj_fun=obj_fun,
                                obj_type='MIN')
        self.add(obj)




def test1():
    from metku.sections.steel.catalogue import ipe_profiles, shs_profiles, \
        hea_profiles, rhs_profiles
    import metku.truss2d as t2d
    import metku.frame2d as f2d

    simple = dict(
        H0=0,
        H1=1800,
        H2=2400,
        L1=10000,
        dx=400,
        n=16
    )

    truss = t2d.Truss2D(simple=simple)

    # Kuormat
    for tc in truss.top_chords:
        truss.add(f2d.LineLoad(tc, [-25, -25], 'y'))

    # Tuet
    truss.add(f2d.XYHingedSupport([0, truss.H0 + truss.H1]))
    truss.add(f2d.XYHingedSupport([truss.L, truss.H0 + truss.H1]))
    truss.generate()
    truss.calculate()

    TC_group = {
        'name': 'TopChords',
        'var_type': 'index',
        'value': 10,
        'values': list(rhs_profiles.keys()),
        'property': 'profile',
        'objects': truss.top_chords
    }

    BC_group = {
        'name': 'BottomChords',
        'var_type': 'index',
        'value': 10,
        'values': list(rhs_profiles.keys()),
        'property': 'profile',
        'objects': truss.bottom_chords
    }

    WEB_group = {
        'name': 'Webs',
        'var_type': 'index',
        'value': 0,
        'values': list(rhs_profiles.keys()),
        'property': 'profile',
        'objects': list(truss.webs.values())
    }

    constraints = {
        'flange_class': 1,
        'web_class': 1,
        'deflection_y': 500,
        'joint_strength_constraints': True

    }

    problem = StructuralProblem(name="TEST",
                                structure=truss,
                                var_groups=[TC_group, BC_group, WEB_group],
                                constraints=constraints
                                )
    problem([50, 50, 30])

if __name__ == '__main__':

    test1()

    # from metku.frame2d.frame2d import *
    # from metku.truss2d import Truss2D
    # from metku.optimization.solvers import *
    #
    # n_bays = 1
    # n_storeys = 1
    # L_bay = 24000
    # H_storey = 8000
    #
    # simple_frame = [n_storeys, n_bays, H_storey, L_bay]
    #
    # frame = Frame2D(simple=simple_frame, create_beams=False, supports='fixed')
    #
    # simple_truss = dict(
    #     H0=H_storey,
    #     H1=1800,
    #     H2=2400,
    #     L1=L_bay / 2,
    #     dx=400,
    #     n=16
    # )
    #
    # truss = Truss2D(simple=simple_truss, fem_model=frame.f)
    # frame.add(truss)
    #
    # for tc in truss.top_chords:
    #     frame.add(LineLoad(tc, [-30, -30], 'y'))
    # for bc in truss.bottom_chords:
    #     frame.add(LineLoad(bc, [-1, -1], 'y'))
    # col1, col2 = frame.columns
    # # frame.add(LineLoad(col1, [3.5, 3.5], 'x'))
    # # frame.add(LineLoad(col2, [0.2, 0.2], 'x'))
    # frame.generate()
    # # truss.generate()
    # frame.calculate()
    # # truss.f.draw()
    #
    # TC_group = {
    #     'name': 'TopChords',
    #     'var_type': 'index',
    #     'value': 10,
    #     'values': list(rhs_profiles.keys()),
    #     'property': 'profile',
    #     'objects': truss.top_chords,
    #     'constraints': {
    #         'buckling_y': True,
    #         'buckling_z': False,
    #         'deflection_y': frame.L / 300,
    #         'deflection_x': frame.H / 300
    #     }
    # }
    #
    # BC_group = {
    #     'name': 'BottomChords',
    #     'var_type': 'index',
    #     'value': 10,
    #     'values': list(rhs_profiles.keys()),
    #     'property': 'profile',
    #     'objects': truss.bottom_chords
    # }
    #
    # WEB_group = {
    #     'name': 'Webs',
    #     'var_type': 'index',
    #     'value': 10,
    #     'values': list(rhs_profiles.keys()),
    #     'property': 'profile',
    #     'objects': list(truss.webs.values())
    # }
    #
    # COL_group = {
    #     'name': 'Columns',
    #     'var_type': 'index',
    #     'value': 10,
    #     'values': list(ipe_profiles.keys()),
    #     'property': 'profile',
    #     # 'properties': ['h', 'b', 'tf', 'tw'],
    #     'objects': frame.columns
    # }
    #
    #
    # problem = StructuralProblem(name="TEST",
    #                             structure=frame,
    #                             var_groups=[TC_group, BC_group, WEB_group,
    #                                         COL_group],
    #                             con_groups=[TC_group],
    #                             constraints={
    #                                 'buckling_y': True,
    #                                 'buckling_z': True,
    #                                 'compression_bending_y': True,
    #                                 'compression_bending_z': True,
    #                                 'compression': True,
    #                                 'tension': True,
    #                                 'shear': True,
    #                                 'deflection_y': frame.L / 200,
    #                                 'deflection_x': frame.H / 300
    #                             })
    #
    # # solver = GA(pop_size=50, mut_rate=0.15)
    # # x0 = [1 for var in problem.vars]
    # # fopt, xopt = solver.solve(problem, x0=x0, maxiter=10, plot=True)
    # # problem(xopt)
    # # frame.plot_deflection(100)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 17 2019

EN 1994-1-1 Design of composite steel and concrete structures: Part 1-1 
General rules and rules for buildings

@author: kmela
"""

import math

try:
    from metku.eurocodes.en1992.constants import gammaC, gammaS, alpha_cc, concrete
except:
    from eurocodes.en1992.constants import gammaC, gammaS, alpha_cc, concrete
    from eurocodes.en1994.constants import gammaV
    

S3L_Studs = {19: {"D": 19.0, "A": 9.52, "H": 31.75},
            22: {"D": 22.0, "A": 9.52, "H": 34.925},
            25: {"D": 19.0, "A": 12.7, "H": 41.275},
            }


class ShearStud:
    """ Class for shear connectors """
    
    def __init__(self,stud=S3L_Studs[19],length=105.0,fy=350,fu=450):
        """ Constructor """
        
        self.d = stud["D"]
        self.a = stud["A"]
        self.h = stud["H"]
        self.length = length
        self.fy = fy
        self.fu = fu
        
    def PRd(self,fck,Ecm,verb=False):
        """ Shear force resistance of a stud """
        
        PRd_steel = 0.8*self.fu*math.pi**2*self.d**2/4
        
        if self.length/self.d > 4:
            alpha = 1.0
        else:
            alpha = 0.2*(self.length/self.d+1)
        
        PRd_concrete = 0.29*alpha*self.d**2*math.sqrt(fck*Ecm)/gammaV
        
        if verb:
            print("PRd(steel) = {0:4.2f} [kN]".format(PRd_steel*1e-3))
            print("PRd(concrete) = {0:4.2f} [kN]".format(PRd_concrete*1e-3))
            print("d = {0:4.2f} [MPa]".format(self.d))
            print("fck = {0:4.2f} [MPa]".format(fck))
            print("Ecm = {0:4.2f} [MPa]".format(Ecm))
            print("alpha = {0:4.2f} [MPa]".format(alpha))
            
        
        
        return min(PRd_steel,PRd_concrete)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

"""
Created on Wed Mar 14 19:11:28 2018

EN 1993-1-1 General rules for steel structures

"""

import math

from metku.eurocodes.en1993.constants import gammaM0, gammaM1, gammaM2

#                 a0     a    b      c     d
buckling_curve = {"a0": 0.13, "a": 0.21, "b": 0.34, "c": 0.49, "d": 0.76}


def sway_imperfection(h, m):
    """ Global sway imperfection of 5.3.2 """
    phi0 = 1 / 200

    alpha_h = min(max(2 / 3, 2 / math.sqrt(h)), 1.0)
    alpha_m = math.sqrt(0.5 * (1 + 1 / m))

    phi = phi0 * alpha_h * alpha_m

    return phi


def epsilon(fy):
    """ Computes the epsilon required in various design rules 
        input: fy -- yield strength in MPa
    
    """

    e = math.sqrt(235.0 / fy)
    return e


def internal_part_in_bending(r, e):
    """ Classification of internal parts of cross-sections in bending
    
    EN 1993-1-1, Table 5.2, part 1
    r = c/t
    e = epsilon
    """

    # Initialize class
    C = 4

    if r <= 72.0 * e:
        C = 1
    elif r <= 83.0 * e:
        C = 2
    elif r <= 124.0 * e:
        C = 3

    return C


def internal_part_in_compression(r, e):
    """ Classification of internal parts of cross-sections in compression
    
    EN 1993-1-1, Table 5.2, part 1
    r = c/t
    e = epsilon
    
    """

    # If e > 1, it means that e is actually yield strength
    if e > 1.0:
        e = epsilon(e)

    C = 4

    if r <= 33.0 * e:
        C = 1
    elif r <= 38.0 * e:
        C = 2
    elif r <= 42.0 * e:
        C = 3

    return C


def internal_part_comp_bend(r, e, a, p):
    """ Classification of internal parts of cross-sections in compression and bending

    EN 1993-1-1, Table 5.2, part 1
    r = c/t
    e = epsilon
    a = alpha: location of plastic neutral axis
    p = psi: location of elastic neutral axis
    
    """

    # If e > 1, it means that e is actually yield strength
    if e > 1.0:
        e = epsilon(e)

    C = 4

    # Check first for class 1 or 2
    if a == 0:
        C = 1
    elif a > 0.5:
        if r <= 396.0 * e / (13 * a - 1):
            C = 1
        elif r <= 456.0 * e / (13 * a - 1):
            C = 2
    else:
        if r <= 36.0 * e / a:
            C = 1
        elif r <= 41.5 * e / a:
            C = 2

    if C > 2:
        # Check for class 3
        if p > -1:
            if r <= 42.0 * e / (0.67 + 0.33 * p):
                C = 3
        elif r <= 62.0 * e * (1 - p) * math.sqrt(-p):
            C = 3

    return C


def outstand_part_in_compression(r, e):
    """ Classification of outstand parts of cross-sections in compression

    EN 1993-1-1, Table 5.2, part 2
    r = c/t
    e = epsilon
    
    """

    # If e > 1, it means that e is actually yield strength
    if e > 1.0:
        e = epsilon(e)

    C = 4

    if r <= 9.0 * e:
        C = 1
    elif r <= 10.0 * e:
        C = 2
    elif r <= 14.0 * e:
        C = 3

    return C


def CHS_class(r, e):
    """ Classification circular hollow sections
    
    EN 1993-1-1, Table 5.2, part 2
    r = d/t
    e = epsilon
    
    """

    C = 4

    if r <= 50 * e ** 2:
        C = 1
    elif r <= 70 * e ** 2:
        C = 2
    elif r <= 90 * e ** 2:
        C = 3

    return C


def tension_resistance(A, fy):
    """ EN 1993-1-1, Eq. (6.6) 
    """
    NplRd = fy * A / gammaM0
    return NplRd


def net_section_resistance(Anet, fu):
    """ EN 1993-1-1, Eq (6.7)
    """
    NuRd = 0.9 * Anet * fu / gammaM2
    return NuRd


def compression_resistance(A, fy):
    """ EN 1993-1-1, Eq (6.10) and (6.11)
        input:
            A .. gross cross section in class 1, 2, or 3
                 effective cross section in class 4
    """
    NcRd = A * fy / gammaM0
    return NcRd


def bending_resistance(WRd, fy):
    """ EN 1993-1-1, Eqs. (6.13) to (6.15)
        input:
            WRd .. section modulus
                   plastic modulus for class 1 or 2
                   elastic modulus for class 3
                   effective elastic modulus for class 4                   
    """
    McRd = WRd * fy / gammaM0
    return McRd


def shear_resistance(Ashear, fy):
    """ EN 1993-1-1, Eq. (6.18)
        input:
            Ashear .. shear area
    """
    VcRd = Ashear * fy / math.sqrt(3) / gammaM0
    return VcRd


def slenderness(A, fy, Ncr):
    """ Non-dimensional slenderness
        input:
            Ncr .. elastic buckling force
    """
    return math.sqrt(A * fy / Ncr)


def buckling_reduction_factor(slend, imp_factor=buckling_curve["c"]):
    """ Reduction factor for flexural buckling
        EN 1993-1-1, Eq. (6.49)
    """
    p = 0.5 * (1 + imp_factor * (slend - 0.2) + slend ** 2)
    chi = min(1 / (p + math.sqrt(p ** 2 - slend ** 2)), 1.0)

    return chi


def buckling_strength(A, fy, chi):
    """ EN 1993-1-1, Eqs. (6.47) and (6.48)
    """
    NbRd = chi * A * fy / gammaM1
    return NbRd


# def beam_column_utility(NEd,NRd,chi,MyEd,MyRd,ky,MzEd=0.0,MzRd=1.0,chiLT=1.0):
def beam_column_utility(UN, UMy, ky, chiLT=1.0, UMz=0.0):
    """ EN 1993-1-1, Eqs. (6.61) and (6.62) 
        input:
            UN = NEd/NRd/chi
            UMy = MyEd/MyRd
            UMz = MzEd/MzRd
                
    """
    Ut = UN / gammaM1 + ky * UMy / chiLT / gammaM1 + UMz / gammaM1
    return Ut


def kyy(UN, slend, Cmy=1.0, section_class=2):
    """ EN 1993-1-1 Table B.1, "Method B"
    """
    if section_class < 3:
        ky = Cmy * (1 + min(slend - 0.2, 0.8) * UN / gammaM1)
    else:
        ky = Cmy * (1 + min(slend, 0.6) * UN / gammaM1)

    return ky


def kzy(ky=0.0, UN=0.0, slend=1.0, CmLT=1.0, section_class=2, susceptible_to_torsion=False):
    """ EN 1993-1-1, Table B.1
    """
    if section_class < 3:
        if susceptible_to_torsion:
            if slend < 0.4:
                kzy = min(0.6 + slend, 1 - 0.1 * slend / (CmLT - 0.25) * UN / gammaM1)
            else:
                kzy = 1 - 0.1 * min(slend, 1) / (CmLT - 0.25) * UN / gammaM1
        else:
            kzy = 0.6 * ky
    else:
        if susceptible_to_torsion:
            kzy = 1 - 0.05 * min(slend, 1) / (CmLT - 0.25) * UN / gammaM1
        else:
            kzy = 0.8 * ky

    return kzy


def kzz(UN, slend, Cmz, section_class=2, profile="hollow"):
    """ EN 1993-1-1 Table B.1
    """
    if section_class < 3:
        if profile == "hollow":
            kz = Cmz * (1 + min(slend - 0.2, 0.8) * UN / gammaM1)
        else:
            kz = Cmz * (1 + min(2 * slend - 0.6, 1.4) * UN / gammaM1)
    else:
        kz = Cmz * (1 + min(slend, 0.6) * UN / gammaM1)

    return kz


def kyz(kz, section_class=2):
    """ EN 1993-1-1 Table B.1 and B.2
    """
    if section_class < 3:
        ky = 0.6 * kz
    else:
        ky = kz
    return ky


def equivalent_moment_factor(Mend, Mspan=None, load="uniform"):
    """ EN 1993-1-1, Table B.3
        input:
            Mend .. M[0] .. moment at end 1
                 M[1] .. moment at end 2
            Mspan .. moment at mid-span
                 
            load .. "uniform" or "point"
    """

    def psi(Mend):
        # Ratio of smaller end moment and larger end moment
        if abs(Mend[1]) > 1e-6:
            r = Mend[0] / Mend[1]
        else:
            # If one of the end moments is zero, the ratio is automatically
            # zero, too.
            r = 0

        if abs(r) > 1.0:
            r = 1 / r

        return r

    if Mspan is None:
        # If no midspan moment is given, assume that the moment
        # distribution is linear.
        p = psi(Mend)

        # For linear distribution, the Cm factor is not dependent on
        # loading type.
        Cm = max(0.6 + 0.4 * p, 0.4)
    else:
        p = psi(Mend)
        # Mid-span-moment is given. Now the task is to determine
        # the shape of the bending moment diagram based on the
        # three moment values.
        if abs(Mspan - 0.5 * sum(Mend)) < 1e-6:
            # Linear distribution            
            Cm = max(0.6 + 0.4 * p, 0.4)
        else:
            # Nonlinear distribution

            if abs(Mend[0]) > abs(Mend[1]):
                Mh = Mend[0]
            else:
                Mh = Mend[1]

            if abs(Mh) < 1e-6:
                # Both end moments are zero in this case
                alpha = 0
            else:
                alpha = Mspan / Mh

            if abs(alpha) > 1.0 or alpha == 0:
                # Moment in the span is larger than the maximum end moment
                if abs(alpha) > 1.0:
                    alpha = 1 / alpha

                if alpha >= 0:
                    if load == "uniform":
                        Cm = 0.95 + 0.05 * alpha
                    else:
                        Cm = 0.90 + 0.1 * alpha
                else:
                    if p >= 0:
                        if load == "uniform":
                            Cm = 0.95 + 0.05 * alpha
                        else:
                            Cm = 0.90 + 0.1 * alpha
                    else:
                        if load == "uniform":
                            Cm = 0.95 + 0.05 * alpha * (1 + 2 * p)
                        else:
                            Cm = 0.90 + 0.1 * alpha * (1 + 2 * p)
            else:
                # Moment in span is smaller than the maximum end moment
                if alpha >= 0:
                    Cm = max(0.2 + 0.8 * alpha, 0.4)
                else:
                    if p >= 0:
                        if load == "uniform":
                            Cm = max(0.1 - 0.8 * alpha, 0.4)
                        else:
                            Cm = max(-0.8 * alpha, 0.4)
                    else:
                        if load == "uniform":
                            Cm = max(0.1 * (1 - p) - 0.8 * alpha, 0.4)
                        else:
                            Cm = max(0.2 * (-p) - 0.8 * alpha, 0.4)

    return Cm



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
import numpy as np

from metku.optimization.structopt import OptimizationProblem, Variable, \
    NonLinearConstraint, ObjectiveFunction


class ConstrainedRosenbrockProblem(OptimizationProblem):
    """
    https://en.wikipedia.org/wiki/Test_functions_for_optimization
    Rosenbrock function constrained with a cubic and a line
    """

    def __init__(self):
        super().__init__(name="Constrained Rosenbrock")

        self.prob_type = 'continuous'
        self.create_variables()
        self.create_constraints()
        self.create_objective()

    def create_variables(self):
        self.vars = []

        x1 = Variable(name="x1",
                      lb=-1.5,
                      ub=1.5)
        x2 = Variable(name="x2",
                      lb=-0.5,
                      ub=2.5)

        self.vars = [x1, x2]

    def create_constraints(self):
        self.cons = []

        def confun1(x):
            x1, x2 = x

            return (x1 - 1) ** 3 - x2 + 1

        def confun2(x):
            x1, x2 = x

            return x1 + x2 - 2

        con1 = NonLinearConstraint(con_fun=confun1,
                                   name="Constraint 1",
                                   con_type="<",
                                   problem=self)

        con2 = NonLinearConstraint(con_fun=confun2,
                                   name="Constraint 2",
                                   con_type="<",
                                   problem=self)

        self.cons = [con1, con2]

    def create_objective(self):
        def obj_fun(x):
            x1, x2 = x
            return (1 - x1) ** 2 + 100 * (x2 - x1 ** 2) ** 2

        obj = ObjectiveFunction(name="Objective ",
                                obj_fun=obj_fun)
        obj.problem = self

        self.obj = obj


class ConstrainedTownsendProblem(OptimizationProblem):
    """
    https: // en.wikipedia.org / wiki / Test_functions_for_optimization
    Townsend function (modified)
    """
    def __init__(self):
        super().__init__(name="Constrained Townsend")

        self.prob_type = 'continuous'
        self.create_variables()
        self.create_constraints()
        self.create_objective()

    def create_variables(self):
        self.vars = []

        x1 = Variable(name="x1",
                      lb=-2.25,
                      ub=2.5)
        x2 = Variable(name="x2",
                      lb=-2.5,
                      ub=1.75)

        self.vars = [x1, x2]

    def create_constraints(self):
        self.cons = []

        def confun(x):
            x1, x2 = x
            t = np.arctan2(x1, x2)

            b1 = (2 * np.cos(t) - 0.5 * np.cos(2 * t) - 0.25 * np.cos(
                3 * t) - 0.125 * np.cos(4 * t)) ** 2
            b2 = (2 * np.sin(t)) ** 2

            g = x1 ** 2 + x2 ** 2 - b1 - b2
            return g

        constr = NonLinearConstraint(con_fun=confun,
                                     name="Constraint",
                                     con_type='<',
                                     problem=self)

        self.cons = [constr]

    def create_objective(self):
        def obj_fun(x):
            x1, x2 = x

            val = -(np.cos((x1 - 0.1) * x2)) ** 2 - x1 * np.sin(3 * x1 + x2)
            return val

        obj = ObjectiveFunction(name="Objective ",
                                obj_fun=obj_fun)
        obj.problem = self

        self.obj = obj

if __name__ == '__main__':
    from metku.optimization.solvers import *

    problem = ConstrainedRosenbrockProblem()
    solver = SLP()
    x0 = [1,1]
    problem.substitute_variables(x0)
    problem.vars[0].lock()
    x0 = [var.ub for var in problem.vars]

    fopt, xopt = solver.solve(problem, x0=x0, maxiter=50)
    problem(xopt)
